# Copyright (C) 2022 Jae-Won Chung <jwnchung@umich.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes that enclose learning rate scaling rules."""

import math
from dataclasses import dataclass


@dataclass
class SquareRootScaler:
    """Square root scaling.

    Args:
        bs: The initial batch size
        lr: The initial learning rate
    """

    bs: int
    lr: float

    def compute_lr(self, new_bs: int) -> float:
        """Compute the scaled learning rate given the new batch size."""
        return self.lr * math.sqrt(new_bs / self.bs)


@dataclass
class LinearScaler:
    """Linear scaling.

    Args:
        bs: The initial batch size
        lr: The initial learning rate
    """

    bs: int
    lr: float

    def compute_lr(self, new_bs: int) -> float:
        """Compute the scaled learning rate given the new batch size."""
        return self.lr * new_bs / self.bs



# Copyright (C) 2022 Jae-Won Chung <jwnchung@umich.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Multi-Armed Bandit implementations."""

from __future__ import annotations

import warnings

import numpy as np


class GaussianTS:
    """Thompson Sampling policy for Gaussian bandits.

    For each arm, the reward is modeled as a Gaussian distribution with
    known precision. The conjugate priors are also Gaussian distributions.
    """

    def __init__(
        self,
        arms: list[int],
        reward_precision: list[float] | float,
        prior_mean: float = 0.0,
        prior_precision: float = 0.0,
        num_exploration: int = 1,
        seed: int = 123456,
        verbose: bool = True,
    ) -> None:
        """Initialze the object.

        Args:
            arms: Bandit arm values to use.
            reward_precision: Precision (inverse variance) of the reward distribution.
                Pass in a list of `float`s to set the reward precision differently for
                each arm.
            prior_mean: Mean of the belief prior distribution.
            prior_precision: Precision of the belief prior distribution.
            num_exploration: How many static explorations to run when no observations
                are available.
            seed: The random seed to use.
            verbose: Whether to print out what's going on.
        """
        self.name = "GaussianTS"

        self.arms = arms
        self.prior_mean = prior_mean
        self.prior_prec = prior_precision
        self.num_exploration = num_exploration
        self.seed = seed
        self.rng = np.random.default_rng(seed)
        self.verbose = verbose

        # Set the precision of the reward distribution of each arm.
        if isinstance(reward_precision, list):
            self.arm_reward_prec = dict(zip(arms, reward_precision))
        else:
            self.arm_reward_prec = {arm: reward_precision for arm in arms}

        # Initialze the parameter distribution with the prior parameters.
        self.arm_param_mean = dict.fromkeys(arms, prior_mean)
        self.arm_param_prec = dict.fromkeys(arms, prior_precision)

        # Track how many times an arm reward has been observed.
        self.arm_num_observations = dict.fromkeys(arms, 0)

    def fit(
        self,
        decisions: list[int] | np.ndarray,
        rewards: list[float] | np.ndarray,
        reset: bool,
    ) -> None:
        """Fit the bandit on the given list of observations.

        Args:
            decisions: A list of arms chosen.
            rewards: A list of rewards that resulted from choosing the arms in `decisions`.
        """
        decisions_arr = np.array(decisions)
        rewards_arr = np.array(rewards)

        # Fit all arms.
        for arm in self.arms:
            self.fit_arm(arm, rewards_arr[decisions_arr == arm], reset)

    def fit_arm(self, arm: int, rewards: np.ndarray, reset: bool) -> None:
        """Update the parameter distribution for one arm.

        Reference: <https://en.wikipedia.org/wiki/Conjugate_prior>

        Args:
            arm: Arm to fit.
            rewards: Array of rewards observed by pulling that arm.
            reset: Whether to reset the parameters of the arm before fitting.
        """
        if reset:
            self.arm_param_mean[arm] = self.prior_mean
            self.arm_param_prec[arm] = self.prior_prec
            self.arm_num_observations[arm] = 0

        if len(rewards) == 0:
            return

        # Read previous state.
        reward_prec = self.arm_reward_prec[arm]
        mean = self.arm_param_mean[arm]
        prec = self.arm_param_prec[arm]

        # Compute the parameters of the posterior distribution.
        # The reward distribution's precision is given as infinite only when we
        # have exactly one observation for the arm, s.t. sampling yields that
        # exact observation.
        if reward_prec == np.inf:
            new_prec = np.inf
            new_mean = rewards.mean()
        else:
            new_prec = prec + len(rewards) * reward_prec
            new_mean = (prec * mean + reward_prec * rewards.sum()) / new_prec

        # Update state.
        self.arm_param_mean[arm] = new_mean
        self.arm_param_prec[arm] = new_prec
        self.arm_num_observations[arm] += len(rewards)

    def predict(self) -> int:
        """Return the arm with the largest sampled expected reward."""
        # Exploration-only phase.
        # Order is random considering concurrent bandit scenarios.
        arrms = np.array(self.arms)
        for arm in self.rng.choice(arrms, len(arrms), replace=False):
            if self.arm_num_observations[arm] < self.num_exploration:
                if self.verbose:
                    print(f"[{self.name}] Explore arm {arm}.")
                return arm

        # Thomopson Sampling phase.
        expectations = self.predict_expectations()
        if self.verbose:
            print(f"[{self.name}] Sampled mean rewards:")
            for arm, sample in expectations.items():
                print(
                    f"[{self.name}] Arm {arm:4d}: mu ~ N({self.arm_param_mean[arm]:.2f}, "
                    f"{1/self.arm_param_prec[arm]:.2f}) -> {sample:.2f}"
                )
        return max(expectations, key=expectations.get)  # type: ignore

    def predict_expectations(self) -> dict[int, float]:
        """Sample the expected reward for each arm.

        Assumes that each arm has been explored at least once. Otherwise,
        a value will be sampled from the prior.

        Returns:
            A mapping from every arm to their sampled expected reward.
        """
        expectations = {}
        for arm in self.arms:
            mean = self.arm_param_mean[arm]
            prec = self.arm_param_prec[arm]
            if prec == self.prior_prec:
                warnings.warn(f"predict_expectations called when arm '{arm}' is cold.")
            expectations[arm] = self.rng.normal(mean, np.sqrt(np.reciprocal(prec)))
        return expectations



# Copyright (C) 2022 Jae-Won Chung <jwnchung@umich.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the energy-time cost metric function."""

from __future__ import annotations


def zeus_cost(energy: float, time: float, eta_knob: float, max_power: int) -> float:
    """Compute Zeus's energy-time cost metric.

    Trades off ETA and TTA based on the value of `eta_knob`.
    The caller is expected to do bound checking for `eta_knob`,
    because `eta_knob` does not change frequently.

    Args:
        energy: Joules
        time: seconds
        eta_knob: Real number in [0, 1].
        max_power: The maximum power limit of the GPU.

    Returns:
        The cost of the DL training job.
    """
    return eta_knob * energy + (1 - eta_knob) * max_power * time


class ZeusCostThresholdExceededException(Exception):
    """Raised when the predicted cost of the next epoch exceeds the cost threshold.

    This exception is used for terminating all the processes when doing data
    parallel training with multiple processes, because ONLY the master
    process will predict `next_cost` and do the threshold checking. However,
    once the predicted cost exceeds the threshold, we want to terminate ALL
    the processes. Currently this is achieved by throwing an exception at the
    master process. The lauching script will terminate all the processes that
    are still alive.

    Attributes:
        time_consumed (float): Time consumed until the current epoch.
        energy_consumed (float): Energy consumed until the current epoch.
        cost (float): Computed Zeus's energy-time cost metric until the current epoch.
        next_cost (float): Predicted Zeus's energy-time cost metric after next epoch.
        cost_thresh (float): The cost threshold.
    """

    def __init__(
        self,
        time_consumed: float,
        energy_consumed: float,
        cost: float,
        next_cost: float,
        cost_thresh: float,
    ) -> None:
        """Initialize the exception."""
        msg = (
            f"Next expected cost {next_cost:.2f} exceeds cost threshold {cost_thresh:.2f}! "
            f"Stopping. Saved training results: time={time_consumed:.2f}, "
            f"energy={energy_consumed:.2f}, cost={cost:.2f}, reached=false"
        )
        super().__init__(msg)
        self.time_consumed = time_consumed
        self.energy_consumed = energy_consumed
        self.cost = cost
        self.next_cost = next_cost
        self.cost_thresh = cost_thresh



# Copyright (C) 2022 Jae-Won Chung <jwnchung@umich.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the ProfileDataLoader class."""

from __future__ import annotations

import atexit
import os
import subprocess
import time
from typing import Literal

import torch
import torch.distributed as dist
from torch.utils.data import DataLoader, Subset

from zeus.util import get_env


class ProfileDataLoader(DataLoader):
    """A DataLoader class that profiles power and time.

    `ProfileDataLoader` acts just like an ordinary
    [`DataLoader`][torch.utils.data.DataLoader] while profiling power
    consumption and epoch latency under the hood. Power profiling is done by
    spawning the Zeus power monitor as a subprocess. The latency of each epoch
    will be printed out to stdout.

    `ProfileDataLoader` interfaces with the outside world with environment variables.

    - `ZEUS_LOG_PREFIX`      : Prefix for power and time log files.
                               Power log: `f"{log_prefix}+gpu{index}.power.csv"`
                               Time log : `f"{log_prefix}.time.csv"`
    - `ZEUS_MONITOR_PATH`    : Path to the Zeus power monitor.
                               (Default: `"/workspace/zeus/zeus_monitor/zeus_monitor"`)
    - `ZEUS_MONITOR_SLEEP_MS`: How many milliseconds to sleep after measuring power.
                               This is passed to the monitor. (Default: `"100"`)

    `ProfileDataLoader` supports training on only a subset of the dataset and
    scaling time measurements as if trained on the entire dataset.

    `ProfileDataLoader` will assume that training is happening on all GPUs visible
    and spawn one Zeus power monitor process for each GPU. If this is not what you
    want, set `CUDA_VISIBLE_DEVICES` or spawn a Docker container that only mounts
    the GPUs that you would like to use.
    """

    # Power monitor processes
    monitor: list[subprocess.Popen] | None = None

    def __init__(
        self,
        *args,
        batch_size: int,
        split: Literal["train", "eval"],
        subset_proportion: float = 1.0,
        eat_batch_size: bool = False,
        only_scale_time: bool = False,
        **kwargs,
    ) -> None:
        """Initialize the ProfileDataLoader.

        Args:
            batch_size: Batch size to use.
            split: Dataset split. Used when printing out epoch latency.
            subset_proportion: Should be between 0.0 and 1.0. When specified,
                only that proportion of the dataset will be used and the dataloader will
                stop early. Then, the measured epoch latency will be scaled as if the
                whole datset was used.
            only_scale_time: If True, the whole dataset will be used for training, but
                the measured epoch latency will still be scaled based on the value of
                `subset_proportion`. This is useful when you already manually scaled down
                the size of an existing dataset, but still want to simulate training the
                original large dataset.
            eat_batch_size: If True, does not pass the `batch_size` argument to the
                constructor of DataLoader. You won't usually need this.
        """
        # Assumes one epoch per invocation of __iter__.
        self.epoch = 0
        if split not in ["train", "eval"]:
            raise ValueError("split should be either 'train' or 'eval'.")
        self.split = split
        self.scaling_factor = 1.0
        self.start1 = None
        self.start2 = None

        # Retrieve environment variables needed.
        self.monitor_path = get_env("ZEUS_MONITOR_PATH", str)
        self.monitor_sleep_ms = get_env("ZEUS_MONITOR_SLEEP_MS", int, default=100)
        self.log_prefix = get_env("ZEUS_LOG_PREFIX", str)

        # Check if the Zeus power monitor is executable.
        if not os.access(self.monitor_path, os.X_OK):
            raise RuntimeError(f"'{self.monitor_path}' is not executable")

        # Create time.csv and write header.
        if self.split == "train":
            ProfileDataLoader.time_file = open(self.log_prefix + ".time.csv", "w")
            self.time_file.write("epoch,split,time\n")
            self.time_file.flush()

        # Slice out subset of dataset if subset_proportion is given.
        dataset = kwargs["dataset"] if "dataset" in kwargs else args[0]
        if subset_proportion > 1.0 or subset_proportion <= 0.0:
            raise ValueError("subset_proportion should be > 0.0 and <= 1.0.")
        if subset_proportion < 1.0:
            subset_indices = list(range(0, len(dataset), round(1 / subset_proportion)))  # type: ignore
            # See note in __next__ for more about scaling.
            self.scaling_factor = len(dataset) / (len(subset_indices) - batch_size)
            if not only_scale_time:
                subset = Subset(dataset, subset_indices)
                if "dataset" in kwargs:
                    kwargs["dataset"] = subset
                else:
                    args = (subset, *args[1:])

        # Call the constructor of DataLoader.
        if eat_batch_size:
            super().__init__(*args, **kwargs)  # type: ignore
        else:
            super().__init__(*args, batch_size=batch_size, **kwargs)  # type: ignore

    def __iter__(self):
        """Wrap the original `__iter__`, but with power profiling."""
        # On the first epoch, start the Zeus power monitors for each GPU.
        if ProfileDataLoader.monitor is None:
            if (count := torch.cuda.device_count()) == 1 or dist.get_rank() == 0:
                ProfileDataLoader.monitor = []
                for index in range(count):
                    monitor_cmd = [
                        self.monitor_path,
                        self.log_prefix + f"+gpu{index}.power.csv",  # Power log file
                        "0",  # Duration
                        str(self.monitor_sleep_ms),  # Monitor sleep time (ms)
                        str(index),  # GPU_id
                    ]
                    print(f"Launching Zeus monitor {index}...")
                    ProfileDataLoader.monitor.append(
                        subprocess.Popen(
                            monitor_cmd,
                            stdin=subprocess.DEVNULL,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                        )
                    )
                atexit.register(kill_monitor)
        # pylint: disable=attribute-defined-outside-init
        self.iter = super().__iter__()
        self.epoch += 1
        self.start1 = None
        self.start2 = None
        return self

    def __next__(self):
        """Wrap the original `__next__`, but with power profiling."""
        try:
            # Special treatment for the first batch.
            # Data loading takes significantly more time for the first batch. Thus, if we
            # simply measure the first ~ last batch latency of the subset of the dataset
            # and multiply 1/subset_proportion, we end up overestimating time_per_epoch.
            # Thus, we isolate the processing time of the first batch (start2 - start1),
            # scale up second ~ last batch latency with the adjusted scaling factor, and
            # later add the processing time of the first batch.
            #
            # Strange if nest to make the common case number of if statement executions 1.
            if self.start2 is None:
                if self.start1 is None:
                    self.start1 = time.time()
                else:
                    self.start2 = time.time()
            return self.iter.__next__()
        except StopIteration:
            end = time.time()
            if self.start1 and self.start2:
                scaled_time = (
                    self.scaling_factor * (end - self.start2)
                    + self.start2
                    - self.start1
                )
                if torch.cuda.device_count() == 1 or dist.get_rank() == 0:
                    self.time_file.write(f"{self.epoch},{self.split},{scaled_time}\n")
                    self.time_file.flush()
                    print(
                        f"epoch {self.epoch} {self.split} time consumed: {scaled_time:.2f}s"
                    )
            raise


def kill_monitor():
    """Kill all Zeus power monitors."""
    monitor = ProfileDataLoader.monitor
    if monitor is not None:
        for i, proc in enumerate(monitor):
            proc.kill()
            print(f"Stopped Zeus monitor {i}.")



# Copyright (C) 2022 Jae-Won Chung <jwnchung@umich.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abstract classes for implementing custom optimization policies."""

from __future__ import annotations

from abc import ABC, abstractmethod

from zeus.job import Job


class BatchSizeOptimizer(ABC):
    """Finds out the best batch size to use for the job."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the batch size optimizer."""

    @abstractmethod
    def register_job(self, job: Job, batch_sizes: list[int]) -> None:
        """Prepare internal state so that it can handle the given job.

        It is assumed that the state of each [`Job`][zeus.job.Job] will be
        managed separately. Note that [`Job`][zeus.job.Job] is hashable,
        and thus can be used as dictionary keys.

        Args:
            job: New jobs to register.
            batch_sizes: Batch sizes to consider.
        """

    @abstractmethod
    def predict(self, job: Job) -> int:
        """Return the best batch size to use for the job.

        Args:
            job: The job to pick the best batch size for.
        """

    @abstractmethod
    def observe(
        self, job: Job, batch_size: int, cost: float, converged: bool | None = None
    ) -> None:
        """Observe the cost of using the given batch size for the job.

        Args:
            job: The job from which this cost observation resulted.
            batch_size: The batch size used for this run of the job.
            cost: The energy-time cost of running the job.
            converged: Whether the job reached its target metric. If may not have
                reached its target if the job was early stopped based on cost or
                the maximum epoch was reached. For BSO's that do not take this into
                account, `None` can be passed.
        """

    def _log(self, message: str) -> None:
        """Log message with object name."""
        print(f"[{self.name}] {message}")


class PowerLimitOptimizer(ABC):
    """Finds out the best power limit to use for the job and batch size."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the power limit optimizer."""

    @abstractmethod
    def predict(self, job: Job, batch_size: int) -> int | None:
        """Return the best power limit for the job and batch size.

        Args:
            job: The job to pick the best power limit for.
            batch_size: The batch size chosen by the
                [`BatchSizeOptimizer`][zeus.policy.BatchSizeOptimizer] for this job.

        Returns:
            The best power limit, or `None` if profiling results via
            [`observe`][zeus.policy.interface.PowerLimitOptimizer.observe] are needed.
        """

    @abstractmethod
    def observe(self, job: Job, batch_size: int, power_limit: int, cost: float) -> None:
        """Observe the cost of using the given batch size and power limit for the job.

        Args:
            job: The job from which this cost observation resulted.
            batch_size: The batch size used for this run of the job.
            power_limit: The power limit used for this run of the job.
            cost: The cost of running the job.
        """

    def _log(self, message: str) -> None:
        """Log message with object name."""
        print(f"[{self.name}] {message}")



# Copyright (C) 2022 Jae-Won Chung <jwnchung@umich.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the ZeusDataLoader class."""

from __future__ import annotations

import atexit
import json
import os
import signal
import subprocess
import time
import logging
from functools import cached_property
from pathlib import Path
from typing import Generator, Literal
import numpy as np

import pynvml
import torch
import torch.distributed as dist
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler

from zeus import analyze
from zeus.util.check import get_env
from zeus.util.metric import ZeusCostThresholdExceededException, zeus_cost

# JIT profiling states
NOT_PROFILING = "NOT_PROFILING"
WARMING_UP = "WARMING_UP"
PROFILING = "PROFILING"

# Config logging
LOG = logging.Logger(__name__)
LOG.setLevel(logging.INFO)
LOG_HANDLER = logging.StreamHandler()
LOG_FORMATTER = logging.Formatter("%(asctime)s %(message)s")
LOG_HANDLER.setFormatter(LOG_FORMATTER)
LOG.addHandler(LOG_HANDLER)


class ZeusDataLoader(DataLoader):
    r"""Profiles and optimizes GPU power limit.

    `ZeusDataLoader` is integrated into the DNN training script, and transparently
    profiles power and time consumption to determine the optimal GPU power limit.

    # Integration examples

    ## Single-GPU

    ```python
    from zeus.run import ZeusDataLoader

    # The one instantiated with max_epochs becomes the train dataloader
    train_loader = ZeusDataLoader(train_set, batch_size=256, max_epochs=100)
    eval_loader = ZeusDataLoader(eval_set, batch_size=256)

    for epoch_number in train_loader.epochs():
        for batch in train_loader:
            # Learn from batch
        for batch in eval_loader:
            # Evaluate on batch

        train_loader.report_metric(validation_metric)
    ```

    ## Data parallel with multi-GPU on a single-node

    !!! Important
        Zeus assumes that exactly one process manages one GPU, and hence
        one instance of [`ZeusDataLoader`][zeus.run.ZeusDataLoader] exists
        for each GPU.

    Users can integrate Zeus into existing data parallel training scripts
    with five specific steps, which are noted below in the comments.

    Please refer to
    [our integration example with ImageNet](https://github.com/SymbioticLab/Zeus/tree/master/examples/imagenet/train.py)
    for a complete example.

    ```python
    import torch
    import torch.distributed as dist
    import torchvision

    from zeus.run import ZeusDataLoader

    # Step 1: Initialize the default process group.
    # This should be done before instantiating `ZeusDataLoader`.
    dist.init_process_group(
        backend=args.dist_backend,
        init_method=args.dist_url,
    )

    # Step 2: Create a model and wrap it with `DistributedDataParallel`.
    model = torchvision.models.resnet18()
    torch.cuda.set_device(local_rank)
    model.cuda(local_rank)
    # Zeus assumes that exactly one process manages one GPU. If you are doing data
    # parallel training, please use `DistributedDataParallel` for model replication
    # and specify the `device_ids` and `output_device` as below:
    model = torch.nn.parallel.DistributedDataParallel(
        model,
        device_ids=[local_rank],
        output_device=local_rank,
    )

    # Step 3: Create instances of `DistributedSampler` to partition the dataset
    # across the GPUs.
    train_sampler = torch.utils.data.distributed.DistributedSampler(train_set)
    eval_sampler = torch.utils.data.distributed.DistributedSampler(eval_set)

    # Step 4: Instantiate `ZeusDataLoader`.
    # `distributed="dp"` tells `ZeusDataLoader` to operate in data parallel mode.
    # The one instantiated with `max_epochs` becomes the train dataloader.
    train_loader = ZeusDataLoader(train_set, batch_size=256, max_epochs=100,
                                  sampler=train_sampler, distributed="dp")
    eval_loader = ZeusDataLoader(eval_set, batch_size=256, sampler=eval_sampler,
                                 distributed="dp")

    # Step 5: Training loop.
    # Use the train dataloader's `epochs` generator to allow Zeus to early-stop
    # based on the cost. Use `report_metric` to let Zeus know the current
    # validation metric.
    for epoch_number in train_loader.epochs():
        for batch in train_loader:
            # Learn from batch
        for batch in eval_loader:
            # Evaluate on batch

        # Make sure you all-reduce the validation metric across all GPUs,
        # since Zeus expects the final validation metric.
        val_metric_tensor = torch.tensor([validation_metric], device="cuda")
        dist.all_reduce(val_metric_tensor, async_op=False)
        train_loader.report_metric(val_metric_tensor.item())
    ```

    # Environment variables

    `ZeusDataLoader` interfaces with the outside world via environment variables.
    Thus, while `ZeusDataLoader` is paired together with
    [`ZeusMaster`][zeus.run.ZeusMaster] in example scripts, any other "driver" can
    use `ZeusDataLoader` as long as it sets appropriate environment variables.

      - `ZEUS_TARGET_METRIC` : Required. Zeus will stop training when this target
                               validation metric is reached. Will be cast to float.
      - `ZEUS_LOG_DIR`       : Directory to store profiling logs. (Default:` "zeus_log"`)
      - `ZEUS_JOB_ID`        : String to prefix in logs. (Default:` "zeus"`)
      - `ZEUS_COST_THRESH`   : Stop training when the energy-time cost will exceed
                               this threshold.  (Default:` "inf"`)
      - `ZEUS_ETA_KNOB`      : $\eta$ knob to tradeoff between energy and time.
                               Larger values reduce more energy and sacrifice time.
                               (Default:` "0.5"`)
      - `ZEUS_MONITOR_PATH`  : Path to the Zeus power monitor binary.
                               (Default:` "/workspace/zeus/zeus_monitor/zeus_monitor"`)
      - `ZEUS_PROFILE_PARAMS`: Warmup and measure iterations for each power limit,
                               separated by a comma. (Default:` "10,40"`)
      - `ZEUS_USE_OPTIMAL_PL`: Whether to actually use the optimal power limit found.
                               Setting this to false is the Observer Mode described
                               in Section 5. (Default:` "True"`)
    """

    # Global power monitor instances.
    monitors: list[subprocess.Popen] = []

    # The power limit currently set for the GPU.
    current_gpu_pl: int = 0

    # Train batch size to be accessed by the eval dataloader.
    train_batch_size: int = 0

    # Length of the eval dataloader. `epochs` in the train dataloader needs this.
    eval_num_samples: int = 0

    # Train-time power profiling result. Maps power limit to avg_power & throughput.
    train_power_result: dict[int, float] = {}
    train_tput_result: dict[int, float] = {}

    # Eval-time power profiling result. Maps power limit to avg_power & throughput.
    eval_power_result: dict[int, float] = {}
    eval_tput_result: dict[int, float] = {}

    # Cost-optimal power limit. Set by the train dataloader after the last power limit
    # was explored.
    optimal_pl: int = 0

    # Train epoch measurements for time/energy accounting.
    train_epoch_time: list[float] = []
    # The master process will record ALL GPUs' energy consumption during training.
    # GPU_i's energy records is `train_epoch_energy[i]`.
    train_epoch_energy: np.ndarray = np.empty(0)

    # Eval-time latency profiling result. Maps power limit to epoch latency.
    eval_epoch_time: list[float] = []
    # The master process will record ALL GPUs' energy consumption during evaluation.
    # GPU_i's energy records is `eval_epoch_energy[i]`.
    eval_epoch_energy: np.ndarray = np.empty(0)

    def __init__(
        self,
        *args,
        batch_size: int,
        max_epochs: int = -1,
        distributed: Literal["dp"] | None = None,
        **kwargs,
    ) -> None:
        """Initialize the dataloader.

        Args:
            batch_size: Batch size to use for training.
            max_epochs: Maximum number of epochs to train. **Specify this parameter only
                to the train data loader.**
            distributed: Distributed strategy to use for training. If training with single GPU,
                this value should be `None`; if training using data parallel with multi-GPU on
                a single node, this value should be `"dp"`.

        Raises:
            ValueError: `max_epochs` is specified when initializing the evaluation dataloader.
            RuntimeError: `torch.distributed` package is not available.
            RuntimeError: The default process group is not initialized. Make sure to call
                `torch.distributed.init_process_group` to initialize the default process
                group before doing a multiprocessing distributed training.
            ValueError: `self.sampler` is not an instance of `DistributedSampler`. An instance of
                `DistributedSampler` will shuffle and distribute data among GPUs, so it is required
                for data parallel training.
            ValueError: `DistributedSampler` passed in `self.sampler` is inconsistent with the default
                process group. Currently, we assume that all the GPUs in the node will be used for
                training. In this case, the instance of `DistributedSampler` should have
                `sampler.num_replicas == torch.distributed.get_world_size()`
                and `sampler.rank == torch.distributed.get_rank()`.
            TypeError: Parameter `distributed` is not correctly specified. Currently, it can only
                be set as `"dp"` or `None`.
            RuntimeError: Scaling is triggered when the profile window exceeds the number of iterations
                in one epoch. But latter is too small, so scaling can not produce a valid profile window.
                Please consider increasing batch size.
        """
        # Save attributes.
        self.batch_size = batch_size
        self.split = "train" if max_epochs != -1 else "eval"
        self.max_epochs = max_epochs
        self.log_prefix = f"[ZeusDataLoader({self.split})]"

        # Initialize the DataLoader.
        super().__init__(*args, batch_size=batch_size, **kwargs)

        # World size and rank for distributed training.
        # Set default value for single-GPU.
        self.world_size = 1
        self.rank = 0
        # Check whether we are doing a distributed training.
        # Pass in world size and rank.

        self.distributed = distributed

        if self.distributed == "dp":
            # Check if the distributed package is available.
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available.")
            # Check if the process group is initialized.
            if not dist.is_initialized():
                raise RuntimeError(
                    "Default process group has not been initialized,"
                    " please make sure to call `init_process_group`"
                    " before you instantiate `ZeusDataLoader`."
                )
            # Check if `self.sampler` is an instance of DistributedSampler.
            if not isinstance(getattr(self, "sampler", None), DistributedSampler):
                raise ValueError(
                    "Sampler is not an instance of `DistributedSampler`."
                    " Data parallel training on multi-GPU requires a `DistributedSampler`."
                )
            # Check the consistency between the sampler and process group.
            if (
                self.sampler.num_replicas != dist.get_world_size()
                or self.sampler.rank != dist.get_rank()
            ):
                raise ValueError(
                    "`DistributedSampler` is inconsistent with the default process group."
                    f" The default process group has `world_size={dist.get_world_size()}`,"
                    f" `rank={dist.get_rank()}`."
                )
            self.world_size = dist.get_world_size()
            self.rank = dist.get_rank()
        else:
            if self.distributed is not None:
                raise TypeError(
                    '`distributed` currently only accepts `"dp"` or `None`.'
                )

        if self._is_train:
            self._log(
                f"Distributed data parallel: {'ON' if self.world_size > 1 else 'OFF'}"
            )

        if self._is_train:
            if ZeusDataLoader.train_batch_size != 0:
                # If max_epochs is specified when initializing a eval dataloader,
                # it will mistaken itself as a train dataloader.
                # In this case, raise a ValueError.
                raise ValueError("Specify max_epochs only to the train dataloader.")
            # In data parallel training, each DataLoader gets `batch_size=global_batch_size/num_gpus`.
            # So, we scale the `train_batch_size` for the consistency with ZeusMaster.
            # NOTE: Zeus assume `global_batch_size == batch_size * num_gpus`. So please ensure that
            # `global_batch_size` is divisible by `num_gpu` in the training script.
            ZeusDataLoader.train_batch_size = self.batch_size * self.world_size

        # Retrieve environment variables from ZeusMaster.
        self.target_metric = get_env("ZEUS_TARGET_METRIC", float)
        self.logdir = get_env("ZEUS_LOG_DIR", str, default="zeus_log")
        self.job_id = get_env("ZEUS_JOB_ID", str, default="zeus")
        self.cost_thresh = get_env("ZEUS_COST_THRESH", float, default=float("inf"))
        self.eta_knob = get_env("ZEUS_ETA_KNOB", float, default=0.5)
        self.monitor_path = get_env(
            "ZEUS_MONITOR_PATH",
            str,
            default="/workspace/zeus/zeus_monitor/zeus_monitor",
        )
        self.warmup_iter, self.profile_iter = map(
            int, get_env("ZEUS_PROFILE_PARAMS", str, default="10,40").split(",")
        )
        self.use_optimal_pl = get_env("ZEUS_USE_OPTIMAL_PL", bool, default=True)

        # Create ZEUS_LOG_DIR if it does not exist.
        os.makedirs(self.logdir, exist_ok=True)

        # Check whether the monitor path is good.
        if not os.access(self.monitor_path, os.X_OK):
            raise RuntimeError(f"'{self.monitor_path}' is not executable")

        # Whether the target metric was reached.
        self.target_metric_reached = False

        # Construct relevant paths.
        self.train_json = (
            f"{self.logdir}/{self.job_id}+bs{self.train_batch_size}.train.json"
        )
        self.power_json = f"{self.logdir}/bs{self.train_batch_size}.power.json"

        # Numbers related to the dataloader.
        # sample_num: the number of iterations processed in the current epoch.
        # num_samples: the total number of iterations in one epoch.
        self.epoch_num = 0
        self.sample_num = 0
        self.num_samples = len(self)

        # Pass the length of the eval dataloader for `epochs`.
        if not self._is_train:
            ZeusDataLoader.eval_num_samples = self.num_samples

        # If the number of iterations in one epoch (`num_samples`) is smaller than or equal
        # to one profile window (`warmup_iters + profile_iters`), we will not be able to
        # profile for any power limit. So, we scale the profile window to fit in one epoch.
        # We also avoid using the last batch of one epoch, becasue when `drop_last == True`,
        # the last batch will be smaller. This usually happens with large batch size on
        # small datasets, eg. CIFAR100.
        if self._is_train and self.warmup_iter + self.profile_iter >= self.num_samples:
            self._log(
                f"The profile window takes {self.warmup_iter + self.profile_iter}"
                f" iterations ({self.warmup_iter} for warmup + {self.profile_iter}"
                f" for profile) and exceeds the number of iterations ({self.num_samples})"
                f" in one epoch. Scaling the profile window to fit in one epoch..."
            )
            scaling_factor = (self.num_samples - 1) / (
                self.warmup_iter + self.profile_iter
            )
            self.warmup_iter = int(self.warmup_iter * scaling_factor)
            self.profile_iter = int(self.profile_iter * scaling_factor)
            if self.warmup_iter == 0 or self.profile_iter == 0:
                raise RuntimeError(
                    f"Number of iterations in one epoch is {self.num_samples} and"
                    " is too small for applying the scaling. Please consider using"
                    " a smaller batch size. If you are running `run_zeus.py`, please"
                    " pass a smaller value to `--b_max`."
                )
            self._log(
                f"Scaling done! New profile window takes {self.warmup_iter + self.profile_iter}"
                f" iterations ({self.warmup_iter} for warmup + {self.profile_iter} for profile)."
            )

        # Power profiling windows
        #
        # +----- warmup_start (change power limit)
        # |        +----- prof_start (record timestamp)
        # |        |                      +----- prof_end (compute and save results)
        # | warmup |        profile       |
        # v  iter  v          iter        v
        # ================================= =====================
        # |      power limit = 250W       | | power limit = 225W  ...
        # ================================= =====================
        #
        # =======================================================
        # |                         Epoch 1                       ...
        # =======================================================
        #
        # Initialize variables for profiling
        self.warmup_start_sample = 0
        self.prof_start_time = 0.0
        self.prof_start_sample = 0
        self.prof_state = NOT_PROFILING
        self.prof_pl_index = 0

        # Initialize data structure for storing the energy accounting
        # based on the number of GPUs.
        if self._is_train:
            # Sanity check
            assert self.world_size > 0, f"{self.world_size=}"
            assert self.max_epochs > 0, f"{self.max_epochs=}"
            ZeusDataLoader.train_epoch_energy = np.zeros(
                shape=(self.world_size, self.max_epochs), dtype=np.float64
            )
            ZeusDataLoader.eval_epoch_energy = np.zeros(
                shape=(self.world_size, self.max_epochs), dtype=np.float64
            )

        # Initialize NVML and get GPU handle or each GPU at the master process.
        self.gpu_handles = []
        pynvml.nvmlInit()
        for index in range(self.world_size):
            handle = pynvml.nvmlDeviceGetHandleByIndex(index)
            # Set persistent mode.
            pynvml.nvmlDeviceSetPersistenceMode(handle, pynvml.NVML_FEATURE_ENABLED)
            self.gpu_handles.append(handle)
        # Query NVML for the possible power limit range. Unit is mW.
        min_pl, self.max_pl = pynvml.nvmlDeviceGetPowerManagementLimitConstraints(
            self.gpu_handles[0]
        )
        self.power_limits = list(range(self.max_pl, min_pl - 25_000, -25_000))
        if self._is_train:
            self._log(f"Power limit range: {self.power_limits}")

        # Check whether profiling is ON or OFF. If OFF, load the power limit
        # from power_json, and set power limit for all GPUs at the master process.
        if self._is_train and self.rank == 0:
            should_profile = self._should_profile
            self._log(f"Power profiling: {'ON' if should_profile else 'OFF'}")
            # If we need to do profiling, no need to touch the power limit.
            # If profiling is already done, load profile information from power_json.
            # Only then do we have the optimal PL available.
            # We only load in the train dataloader since it populates classvars.
            if not should_profile:
                self._load_power_results()
                self._set_gpu_steady_power_limit()

        # Make sure NVML is shutdown and the monitor is killed when the training script exits.
        if self._is_train:

            def exit_hook():
                pynvml.nvmlShutdown()
                # Master process kills all the monitors.
                if self.rank == 0:
                    for index, monitor in enumerate(self.monitors):
                        monitor.kill()
                        self._log(f"[GPU_{index}] Stopped Zeus monitor.")

            atexit.register(exit_hook)

    def epochs(self) -> Generator[int, None, None]:
        """Yield the current epoch number from 0 until when training should stop.

        Training should stop when

        - the cost reached the cost threshold, or
        - the maximum number of epochs was reached, or
        - the target metric was reached.

        When done, stores the job results in `train_json`.

        Yields:
            Epoch indices starting from zero.

        Raises:
            ZeusCostThresholdExceededException: the predicted cost after the next
                epoch exceeds the cost threshold. When doing data parallel training,
                this exception is used for ternimating all the processes.
        """
        # Sanity check.
        if not self._is_train:
            raise RuntimeError("Use epochs() on the train dataloader.")

        while True:
            # Variables for storing time/energy consumption & cost
            time_consumed, energy_consumed = -1, -1
            cost = -1
            if self.rank == 0:
                # Sanity checks.
                enum = self.epoch_num
                assert (
                    len(self.train_epoch_time) == enum
                ), f"{len(self.train_epoch_time)=}"
                assert (
                    len(self.eval_epoch_time) == enum
                ), f"{len(self.eval_epoch_time)=}"

                # Compute time and energy consumption up to now.
                # Compute time consumption at GPU_0
                time_consumed = sum(self.train_epoch_time + self.eval_epoch_time)
                # Compute energy consumption over all the GPUs
                energy_consumed = (
                    self.train_epoch_energy.sum() + self.eval_epoch_energy.sum()
                )
                cost = zeus_cost(
                    energy_consumed,
                    time_consumed,
                    self.eta_knob,
                    self.max_pl // 1000 * self.world_size,
                )
                self._log(
                    f"Up to epoch {self.epoch_num}: "
                    f"time={time_consumed:.2f}, energy={energy_consumed:.2f}, cost={cost:.2f}"
                )

            # target_metric_reached is set when the current validation metric is reported to
            # the train dataloader after the end of each epoch.
            # Stop if the target metric was reached.
            if self.target_metric_reached:
                if self.rank == 0:
                    # Sanity check that time/energy consumption & cost are valid in master process.
                    assert time_consumed >= 0 and energy_consumed >= 0 and cost >= 0
                    self._log(
                        f"Target metric {self.target_metric} was reached! Stopping."
                    )
                    self._save_train_results(energy_consumed, time_consumed, cost, True)
                return

            # Max epoch is a hard stop.
            if self.epoch_num >= self.max_epochs:
                if self.rank == 0:
                    # Sanity check that time/energy consumption & cost are valid in master process.
                    assert time_consumed >= 0 and energy_consumed >= 0 and cost >= 0
                    self._log(
                        f"Maximum number of epochs {self.max_epochs} reached. Stopping."
                    )
                    self._save_train_results(
                        energy_consumed, time_consumed, cost, False
                    )
                return

            # No need to do anything in the first epoch.
            if self.epoch_num == 0:
                yield 0
                continue

            # Just continue if we're profiling.
            # This will ignore and continue training even if the cost threshold was exceeded.
            # However, the profiling cost actually exceeding the cost threshold would not
            # happen frequently. It's more like a wrong cost threshold.
            if self._should_profile:
                if cost >= self.cost_thresh:
                    self._log(
                        f"{cost=:.2f} exceeded threshold {self.cost_thresh:.2f} at GPU_{self.rank}, "
                        "but just continue since we're profiling."
                    )
                yield self.epoch_num
                continue

            if self.rank == 0:
                # Sanity check that time/energy consumption & cost are valid in master process.
                assert time_consumed >= 0 and energy_consumed >= 0 and cost >= 0

                # We want to predict whether running the next epoch will exceed the cost threshold.
                next_train_time = (
                    self.num_samples / self.train_tput_result[self.optimal_pl]
                )
                next_eval_time = (
                    self.eval_num_samples / self.eval_tput_result[self.optimal_pl]
                )
                next_time = next_train_time + next_eval_time
                next_train_energy = (
                    next_train_time * self.train_power_result[self.optimal_pl]
                )
                next_eval_energy = (
                    next_eval_time * self.eval_power_result[self.optimal_pl]
                )
                next_energy = next_train_energy + next_eval_energy
                self._log(
                    f"Optimal PL train & eval expected time={next_time:.2f} energy={next_energy:.2f}"
                )
                next_time_consumed = time_consumed + next_time
                next_energy_consumed = energy_consumed + next_energy
                next_cost = zeus_cost(
                    next_energy_consumed,
                    next_time_consumed,
                    self.eta_knob,
                    self.max_pl // 1000 * self.world_size,
                )
                self._log(
                    f"Expected next epoch: time={next_time_consumed:.2f}, "
                    f"energy={next_energy_consumed:.2f}, "
                    f"cost={next_cost:.2f}"
                )

                # Stop if the predicted cost of the next epoch exceeds the cost threshold.
                if next_cost >= self.cost_thresh:
                    # Save training results
                    self._save_train_results(
                        energy_consumed, time_consumed, cost, False
                    )
                    # NOTE: We use a customized exception to terminate ALL the processes for
                    # the purpose of multiprocessing management.
                    # When doing data parallel training on multiple processes, ONLY the master
                    # process will predict `next_cost` and do the threshold checking. However,
                    # once the predicted cost exceeds the threshold, we want to terminate ALL
                    # the processes. Currently this is achieved by throwing an exception at the
                    # master process. The lauching script will terminate all the processes that
                    # are still alive.
                    raise ZeusCostThresholdExceededException(
                        time_consumed,
                        energy_consumed,
                        cost,
                        next_cost,
                        self.cost_thresh,
                    )

            yield self.epoch_num

    def report_metric(self, metric: float, higher_is_better: bool) -> None:
        """Report the validation metric to the train dataloader.

        If doing data parallel training, please make sure
        to call `dist.all_reduce()` to reduce the validation metric across all GPUs
        before calling `train_loader.report_metric()`.

        Args:
            metric: The validation metric of the current epoch.
            higher_is_better: For example, this should be `True` for accuracy
                and `False` for error.
        """
        assert self._is_train, "Use report_metric on the train dataloader."
        if higher_is_better:
            if metric >= self.target_metric:
                self.target_metric_reached = True
        else:
            if metric <= self.target_metric:
                self.target_metric_reached = True

    @property
    def _should_profile(self) -> bool:
        """Whether profiling is not done."""
        return not Path(self.power_json).exists()

    @property
    def _power_limits_left(self) -> bool:
        """Whether there are power limits left to profile."""
        return self.prof_pl_index < len(self.power_limits)

    def _compute_optimal_pl(self) -> int:
        """Return the cost-optimal power limit."""
        # Sanity checks.
        assert ZeusDataLoader.train_tput_result
        assert ZeusDataLoader.train_power_result
        # Only compute optimal PL at master process.
        assert self.rank == 0

        # Compute power cost
        tput = ZeusDataLoader.train_tput_result
        power = ZeusDataLoader.train_power_result
        cost_map = {
            pl: (
                self.eta_knob * power[pl]
                + (1 - self.eta_knob) * self.max_pl * self.world_size
            )
            / tput[pl]
            for pl in self.power_limits
        }
        optimal_pl = min(cost_map.keys(), key=cost_map.get)  # type: ignore
        self._log(f"Cost-optimal power limit is {optimal_pl//1000}W")
        return optimal_pl

    def _set_gpu_power_limit(self, power_limit: int) -> None:
        """Set the GPU's power limit using NVML.

        This method only invokes NVML when `power_limit` is not the same as
        the current GPU power limit.

        Args:
            power_limit: Power limit to set.
        """
        # Sanity check.
        # Only set power limit at master process.
        assert self.rank == 0
        assert len(self.gpu_handles) == self.world_size

        # Set power limit for all GPUs.
        if self.current_gpu_pl != power_limit:
            for index in range(self.world_size):
                pynvml.nvmlDeviceSetPowerManagementLimit(
                    self.gpu_handles[index], power_limit
                )
                self._log(f"[GPU_{index}] Set GPU power limit to {power_limit//1000}W.")
            ZeusDataLoader.current_gpu_pl = power_limit

    def _set_gpu_steady_power_limit(self) -> None:
        """Set the steady power limit based on self.use_optimal_pl."""
        # Sanity check.
        # Only set power limit at master process.
        assert self.rank == 0

        power_limit = ZeusDataLoader.optimal_pl if self.use_optimal_pl else self.max_pl
        self._log(
            "Steady state power limit: "
            f"{'OPT' if self.use_optimal_pl else 'MAX'} {power_limit//1000}W"
        )
        self._set_gpu_power_limit(power_limit)

    def _log(
        self, message: str, level: int = logging.INFO, master_only: bool = True
    ) -> None:
        """Print out message with prefix.

        Args:
            message: The message to log out.
            level: The logging level to use. (Default: `logging.INFO`)
            master_only: Whether only logged by master process. Usually set to True for the
                global logging and False for the GPU-specific logging . If set to False,
                a prefix indicates which GPU this log comes from will be included as well.
                (Default: `True`)
        """
        if master_only:
            if self.rank == 0:
                LOG.log(level, "%s %s", self.log_prefix, message)
        else:
            gpu_log_prefix = f"[GPU_{self.rank}]"
            LOG.log(level, "%s %s %s", self.log_prefix, gpu_log_prefix, message)

    @cached_property
    def _is_train(self) -> bool:
        """Return whether this dataloader is for training."""
        return self.split == "train"

    def _power_log_path(self, rank: int) -> str:
        """Build the path for the power monitor log file at the GPU with rank."""
        return f"{self.logdir}/bs{self.train_batch_size}+e{self.epoch_num}+gpu{rank}.power.log"

    def _start_monitor(self) -> None:
        """Start the power monitor subprocess."""
        # Sanity checks.
        assert not ZeusDataLoader.monitors
        # Only the master process starts the monitors.
        assert self.rank == 0

        # Start monitors. Master process starts and records all monitors.
        for index in range(self.world_size):
            monitor = subprocess.Popen(
                [self.monitor_path, self._power_log_path(index), "0", "100"],
            )
            self._log(f"[GPU_{index}] Zeus monitor started.")
            ZeusDataLoader.monitors.append(monitor)

    def _kill_monitor(self) -> None:
        """Kill the power monitor subprocess."""
        # Sanity checks.
        assert ZeusDataLoader.monitors
        # Only the master process kills the monitors.
        assert self.rank == 0

        # Kill monitors.
        for monitor in ZeusDataLoader.monitors:
            monitor.send_signal(signal.SIGINT)
        for monitor in ZeusDataLoader.monitors:
            monitor.wait(timeout=1.0)

        # Cleanup the monitor list
        ZeusDataLoader.monitors = []

    def _start_warmup(self) -> None:
        """Let the GPU run for some time with the poewr limit to profile."""
        # Sanity checks.
        assert self._should_profile, f"start_warmup: {self._should_profile=}"
        assert self._is_train, f"start_warmup: {self._is_train=}"
        assert self._power_limits_left, f"start_warmup: {self._power_limits_left=}"
        # Sanity check that this profile window ends before the end of the current epoch.
        assert (
            self.sample_num + self.warmup_iter + self.profile_iter < self.num_samples
        ), (
            "start_warmup: "
            f"end_of_this_profile_window {self.sample_num + self.warmup_iter + self.profile_iter} "
            f"< end_of_this_epoch {self.num_samples}"
        )

        # Call cudaSynchronize to make sure this is the iteration boundary.
        torch.cuda.synchronize()

        # Change power limit.
        if self.rank == 0:
            power_limit = self.power_limits[self.prof_pl_index]
            self._set_gpu_power_limit(power_limit)

            self._log(f"Warm-up started with power limit {self.current_gpu_pl//1000}W")

        self.warmup_start_sample = self.sample_num

        # Set profiling state.
        self.prof_state = WARMING_UP

    def _start_prof(self) -> None:
        """Start profiling power consumption for the current power limit."""
        # Sanity checks.
        assert self._should_profile, f"start_prof: {self._should_profile=}"
        assert self._is_train, f"start_prof: {self._is_train=}"
        assert self._power_limits_left, f"start_prof: {self._power_limits_left=}"
        # Sanity check that this profile window ends before the end of the current epoch.
        assert self.sample_num + self.profile_iter < self.num_samples, (
            "start_prof: "
            f"end_of_this_profile_window {self.sample_num + self.profile_iter} "
            f"< end_of_this_epoch {self.num_samples}"
        )

        # Start profile timer.
        self.prof_start_time = time.monotonic()

        # Set the sample number when we started profiling.
        self.prof_start_sample = self.sample_num

        # Set profiling state.
        self.prof_state = PROFILING

        self._log(f"Profile started with power limit {self.current_gpu_pl//1000}W")

    def _end_prof(self) -> None:
        """End profiling power consumption for this power limit.

        Raises:
            ValueError: ValueError raised by sklearn.metrics.auc in analyze.avg_power,
                might due to profile window too small. In this case, user should consider
                increasing profile window.
        """
        # Sanity checks.
        assert self._should_profile, f"end_prof: {self._should_profile=}"
        assert self._is_train, f"end_prof: {self._is_train=}"
        assert self._power_limits_left, f"end_prof: {self._power_limits_left=}"
        # Sanity check that this profile window ends before the end of the current epoch.
        assert self.sample_num < self.num_samples, (
            "end_prof: "
            f"end_of_this_profile_window {self.sample_num} "
            f"< end_of_this_epoch {self.num_samples}"
        )

        # Set profiling state.
        self.prof_state = NOT_PROFILING

        # Call cudaSynchronize to make sure this is the iteration boundary.
        torch.cuda.synchronize()

        # Freeze time.
        now = time.monotonic()

        # Advance to the next power limit. Affects self.power_limits_left.
        self.prof_pl_index += 1

        if self.rank == 0:
            # Summing up the average power on all GPUs.
            sum_avg_power = 0
            for index in range(self.world_size):
                # Compute and save average power.
                # The monitor is still running, so we just integrate from the beginning
                # of this profiling window (of course exclude warmup) up to now.
                # The power log file only records for the current epoch,
                # so we compute an offset.
                try:
                    avg_power = analyze.avg_power(
                        self._power_log_path(index),
                        start=self.prof_start_time - self.epoch_start_time,
                    )
                except ValueError:
                    self._log(
                        "ValueError from analyze.avg_power, please consider increasing self.profile_iter.",
                        logging.ERROR,
                    )
                    raise
                sum_avg_power += avg_power
            self.train_power_result[self.current_gpu_pl] = sum_avg_power

            # Compute and save throughput. We use the time at the master process.
            time_consumed = now - self.prof_start_time
            samples_processed = self.sample_num - self.prof_start_sample
            throughput = samples_processed / time_consumed
            self.train_tput_result[self.current_gpu_pl] = throughput

            self._log(f"Profile done with power limit {self.current_gpu_pl//1000}W")

            # If we're done with all power limits, compute the optimal power limit
            # and change to that power limit for the rest of the epoch.
            # This will lead to the eval epoch being run with the optimal power limit,
            # and since self.should_profile is still True, tput/power will be profiled.
            # Profiling the optimal power limit on eval set will help us better predict
            # the time and energy consumed in the next eval epoch, to help us decide
            # whether running next epoch will exceed the cost threshold.
            if not self._power_limits_left:
                self._log("This was the last power limit to explore.")
                ZeusDataLoader.optimal_pl = self._compute_optimal_pl()
                self._set_gpu_power_limit(ZeusDataLoader.optimal_pl)

    def _save_power_results(self) -> None:
        """Write the power profiling results to `power_json`."""
        # Sanity check.
        # Only save power results at master process.
        assert self.rank == 0

        prof_result = dict(
            job_id=self.job_id,  # Not used. Just for the purpose of record.
            train_power=self.train_power_result,
            train_throughput=self.train_tput_result,
            eval_power=self.eval_power_result,
            eval_throughput=self.eval_tput_result,
            optimal_pl=self.optimal_pl,
        )
        # NOTE: Write-then-move needed if we're handling concurrent jobs.
        with open(self.power_json, "w") as f:
            json.dump(prof_result, f)
        with open(self.power_json, "r") as f:
            self._log("Power profiling done.")
            self._log(f"Saved {self.power_json}: {f.read()}")

    def _load_power_results(self) -> None:
        """Load power profiling information into the class from `power_json`."""
        # Sanity check.
        # Only load power results at master process.
        assert self.rank == 0

        # Helper function that casts the keys of a dictionary to integer.
        def as_int_key(dictionary: dict[str, float]) -> dict[int, float]:
            result = {}
            for key, value in dictionary.items():
                result[int(key)] = value
            return result

        with open(self.power_json, "r") as f:
            power_results = json.load(f)

        ZeusDataLoader.train_power_result = as_int_key(power_results["train_power"])
        ZeusDataLoader.train_tput_result = as_int_key(power_results["train_throughput"])
        ZeusDataLoader.eval_power_result = as_int_key(power_results["eval_power"])
        ZeusDataLoader.eval_tput_result = as_int_key(power_results["eval_throughput"])
        ZeusDataLoader.optimal_pl = power_results["optimal_pl"]

        self._log(f"Loaded {self.power_json}: {power_results}")

    def _save_train_results(
        self, energy: float, time_: float, cost: float, reached: bool
    ) -> None:
        """Write the job training results to `train_json`."""
        # Sanity check.
        # Only load power results at master process.
        assert self.rank == 0

        train_result = dict(
            energy=energy,
            time=time_,
            cost=cost,  # Not used. Just for reference.
            num_epochs=self.epoch_num,  # Not used. Just for reference.
            reached=reached,
        )
        with open(self.train_json, "w") as f:
            json.dump(train_result, f)
        with open(self.train_json, "r") as f:
            self._log("Training done.")
            self._log(f"Saved {self.train_json}: {f.read()}")

    # pylint: disable=attribute-defined-outside-init
    def __iter__(self):
        """Signal the beginning of an epoch."""
        # Sanity check that there is no incomplete profile window at the beginning of epoch,
        # because we start profiling only if the entire profiling window can fit in the rest of
        # the training epoch.
        assert self.prof_state == NOT_PROFILING, f"__iter__: {self.prof_state=}"

        # Update counters.
        self.epoch_num += 1
        self.sample_num = 0
        self._log(f"Epoch {self.epoch_num} begin.")

        # Start epoch timer.
        self.epoch_start_time = time.monotonic()

        # Cache the dataloader iterator.
        self.iter = super().__iter__()

        # The power limit of the GPU is only changed by the train dataloader.
        if self._is_train and self.rank == 0:
            # The train loader always starts the monitor, and the eval loader kills it.
            self._start_monitor()
            # If we're not profiling, use the steady state power limit.
            # If we are profiling, the power limit will be set in __next__ with warmup.
            # Power limit result is already loaded in when initializing the train dataloader,
            # so we just set the power limit directly.
            if not self._should_profile:
                self._set_gpu_steady_power_limit()

        return self

    def __next__(self):
        """Signal the beginning of an iteration."""
        # Update counters.
        self.sample_num += 1

        # Try to fetch next batch.
        try:
            data = next(self.iter)
        except StopIteration:
            # End of this epoch.
            # Sanity check that there is no incomplete profile window at the end of epoch.
            assert self.prof_state == NOT_PROFILING, f"__next__: {self.prof_state=}"

            # Make sure all GPU operations are done so that now is the *actual* end of this epoch.
            torch.cuda.synchronize()

            # The eval dataloader kills the monitor.
            if not self._is_train and self.rank == 0:
                self._kill_monitor()

            # Compute epoch time and energy consumption.
            # We're interested in the actual time/energy consumption here.
            #
            #   <----------------- monitor lifetime ------------------>
            #
            #   =======================================================
            #   |                Train                    ||   Eval   |
            #   =======================================================
            #   ^                                         ^^          ^
            #   |                                        / |          |
            #   epoch_start_time          time.monotonic() |          |
            #   for train loader          for train loader |          |
            #                                              |          |
            #                               epoch_start_time   time.monotonic()
            #                                for eval loader   for eval loader
            #
            if self.rank == 0:
                # Sanity check that `epoch_num` is within valid range
                assert self.epoch_num >= 1, f"__next__: {self.epoch_num=}"
                # Compute the time/energy consumption for this epoch.
                time_consumption = time.monotonic() - self.epoch_start_time
                if self._is_train:
                    self.train_epoch_time.append(time_consumption)
                    # Record the energy consumption for each GPU.
                    for index in range(self.world_size):
                        # The monitor is still running, and we integrate over the entire log.
                        energy_consumption = analyze.energy(self._power_log_path(index))
                        self.train_epoch_energy[index][
                            self.epoch_num - 1
                        ] = energy_consumption
                else:
                    # We just killed the monitor. Integrate the last time_consumption seconds.
                    self.eval_epoch_time.append(time_consumption)
                    sum_energy_consumption = 0
                    # Record the energy consumption for each GPU.
                    for index in range(self.world_size):
                        # We set the `start` to exclude the logs during train time.
                        energy_consumption = analyze.energy(
                            self._power_log_path(index), start=-time_consumption
                        )
                        self.eval_epoch_energy[index][
                            self.epoch_num - 1
                        ] = energy_consumption
                        sum_energy_consumption += energy_consumption
                    # For the eval dataloader, we want to record the throughput and power
                    # for the current power limit. Since the train dataloader sets the power limit
                    # to the optimal power limit right after profiling is done, this will naturally
                    # record the tput/power of the optimal PL. From the following epochs where we
                    # don't profile anything, we directly use these values to compute the time and
                    # energy consumed.
                    if self._should_profile:
                        self.eval_tput_result[self.current_gpu_pl] = (
                            self.num_samples / time_consumption
                        )
                        self.eval_power_result[self.current_gpu_pl] = (
                            sum_energy_consumption / time_consumption
                        )
                        # The optimal PL being known means that all power limits have been explored.
                        # Let us end profiling by writing profile information to `power_json`.
                        if self.optimal_pl != 0:
                            self._save_power_results()
                self._log(
                    f"{self.split} epoch {self.epoch_num} done: "
                    f"time={time_consumption:.2f} energy={energy_consumption:.2f}"
                )

            # Re-raise StopIteration.
            raise

        # We're in the middle of an epoch. The train loader has power limits left to profile.
        if self._is_train and self._should_profile and self._power_limits_left:
            # We weren't doing anything. Start warming up if the iterations left in
            # the current epoch can accommodate at least one profile window.
            if (
                self.prof_state == NOT_PROFILING
                and self.sample_num + self.warmup_iter + self.profile_iter
                < self.num_samples
            ):
                self._start_warmup()
            # We're done warming up. Start the actual profiling window.
            elif (
                self.prof_state == WARMING_UP
                and self.sample_num - self.warmup_start_sample == self.warmup_iter
            ):
                self._start_prof()
            # We're done profiling. Stop the profiling window and gather results.
            elif (
                self.prof_state == PROFILING
                and self.sample_num - self.prof_start_sample == self.profile_iter
            ):
                self._end_prof()

        return data



# Copyright (C) 2022 Jae-Won Chung <jwnchung@umich.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for result analysis."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path
from typing import cast

import pandas as pd
from sklearn.metrics import auc


@dataclass
class HistoryEntry:
    """Represents the config and result of a job run that may have failed.

    Attributes:
        bs: Batch size
        pl: Power limit
        energy: Energy consumption in Joules
        reached: Whether the target metric was reached at the end
        time: Time consumption in seconds
    """

    bs: int
    pl: int
    energy: float
    reached: bool
    time: float


def energy(
    logfile: Path | str,
    start: float | None = None,
    end: float | None = None,
) -> float:
    """Compute the energy consumption from the Zeus monitor power log file.

    `start` and `end` are in units of seconds, relative to the beginning of
    the time window captured by the log file. Only the time window between
    `start` and `end` will be considered when computing energy.

    `start` and `end` can be negative, in which case the pointers wrap around
    and effectively the absolute value is subtracted from the end of the window.

    Args:
        logfile: Path to the power log file produced by the Zeus monitor.
        start: Start time of the window to consider.
        end: End time of the window to consider.
    """
    df = cast(pd.DataFrame, pd.read_csv(logfile, engine="python", skipfooter=1))
    df["Time"] = pd.to_datetime(df["Time"])
    start_timestamp = df.iloc[0]["Time"]
    end_timestamp = df.iloc[-1]["Time"]
    if start is not None:
        origin = start_timestamp if start >= 0.0 else end_timestamp
        df = df.loc[df["Time"] >= origin + timedelta(seconds=start)]
    if end is not None:
        origin = start_timestamp if end >= 0.0 else end_timestamp
        df = df.loc[df["Time"] <= origin + timedelta(seconds=end)]
    seconds = _get_seconds(df)
    watts = _get_watts(df)
    return auc(seconds, watts)


def avg_power(
    logfile: Path | str,
    start: float | None = None,
    end: float | None = None,
) -> float:
    """Compute the average power consumption from the Zeus monitor power log file.

    `start` and `end` are in units of seconds, relative to the beginning of
    the time window captured by the log file. Only the time window between
    `start` and `end` will be considered when computing average power.

    `start` and `end` can be negative, in which case the pointers wrap around
    and effectively the absolute value is subtracted from the end of the window.

    Args:
        logfile: Path to the power log file produced by the Zeus monitor.
        start: Start time of the window to consider.
        end: End time of the window to consider.

    Raises:
        ValueError: From `sklearn.metrics.auc`, when the duration of the
            profiling window is too small.
    """
    df = cast(pd.DataFrame, pd.read_csv(logfile, engine="python", skipfooter=1))
    df["Time"] = pd.to_datetime(df["Time"])
    if start is not None:
        df = df.loc[df["Time"] >= df.iloc[0]["Time"] + timedelta(seconds=start)]
    if end is not None:
        df = df.loc[df["Time"] <= df.iloc[0]["Time"] + timedelta(seconds=end)]
    seconds = _get_seconds(df)
    watts = _get_watts(df)
    area = auc(seconds, watts)
    return area / (seconds.max() - seconds.min())


def _get_seconds(df: pd.DataFrame) -> pd.Series:
    return df["Time"].map(lambda t: t.timestamp())


def _get_watts(df: pd.DataFrame) -> pd.Series:
    return df["Power"].div(1000.0)



# Copyright (C) 2022 Jae-Won Chung <jwnchung@umich.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Optimization policies for Zeus.

[`PowerLimitOptimizer`][zeus.policy.interface.PowerLimitOptimizer] and
[`BatchSizeOptimizer`][zeus.policy.interface.BatchSizeOptimizer] are
abstract classes. Users can implement custom policies by extending the
abstract classes and implementing required method.
Currently, [`Simulator`][zeus.simulate.Simulator] supports custom policies
for both classes, while [`ZeusMaster`][zeus.run.ZeusMaster] only supports
custom [`BatchSizeOptimizer`][zeus.policy.interface.BatchSizeOptimizer]s.
Custom [`PowerLimitOptimizer`][zeus.policy.PowerLimitOptimizer]s will
have to be integrated into [`ZeusDataLoader`][zeus.run.ZeusDataLoader].
"""

from zeus.policy.interface import BatchSizeOptimizer, PowerLimitOptimizer
from zeus.policy.optimizer import JITPowerLimitOptimizer, PruningGTSBatchSizeOptimizer



# Copyright (C) 2022 Jae-Won Chung <jwnchung@umich.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the Job specification dataclass."""

from __future__ import annotations

from dataclasses import dataclass, field

import pandas as pd

from zeus.util import LinearScaler, SquareRootScaler


@dataclass(frozen=True, unsafe_hash=True)
class Job:
    """Job specification tuple.

    Attributes:
        dataset: Name of the dataset.
        network: Name of the DNN model.
        optimizer: Name of the optimizer, e.g. Adam.
        target_metric: Target validation metric.
        max_epochs: Maximum number of epochs to train before terminating.
        default_bs: Initial batch size (b0) provided by the user.
        default_lr: Learning rate corresponding to the default batch size.
        workdir: Working directory in which to launch the job command.
        command: Job command template. See [`gen_command`][zeus.job.Job.gen_command].
    """

    dataset: str
    network: str
    optimizer: str
    target_metric: float
    max_epochs: int
    default_bs: int | None = None
    default_lr: float | None = None
    workdir: str | None = None
    command: list[str] | None = field(default=None, hash=False, compare=False)

    def __str__(self) -> str:
        """Generate a more conside representation of the object."""
        return (
            f"Job({self.dataset},{self.network},{self.optimizer},{self.target_metric}"
            f"{f',bs{self.default_bs}' if self.default_bs is not None else ''}~{self.max_epochs})"
        )

    def to_logdir(self) -> str:
        """Generate a logdir name that explains this job."""
        return (
            f"{self.dataset}+{self.network}+bs{self.default_bs}"
            f"+{self.optimizer}+lr{self.default_lr}"
            f"+tm{self.target_metric}+me{self.max_epochs}"
        )

    def filter_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """Pick out the rows corresponding to this job from the DataFrame."""
        return df.loc[
            (df.dataset == self.dataset)
            & (df.network == self.network)
            & (df.optimizer == self.optimizer)
            & (df.target_metric == self.target_metric)
        ]

    def gen_command(
        self,
        batch_size: int,
        learning_rate: float,
        seed: int,
        rec_i: int,
    ) -> list[str]:
        """Format the job command with given arguments.

        Args:
            batch_size: Batch size to use for this job launch.
            learning_rate: Learning rate to use for this job launch.
            seed: Random seed to use for this job launch.
            rec_i: Recurrence number of this job launch.
        """
        assert self.command, "You must provide a command format string for this job."
        command = []
        for piece in self.command:
            if piece in ["{bs}", "{batch_size}"]:
                command.append(str(batch_size))
            elif piece in ["{lr}", "{learning_rate}"]:
                command.append(str(learning_rate))
            elif piece == "{seed}":
                command.append(str(seed))
            elif piece in ["{epoch}", "{epochs}"]:
                command.append(str(self.max_epochs))
            elif piece == "{slice_number}":
                command.append(str(rec_i))
            elif piece == "{target_metric}":
                command.append(str(self.target_metric))
            else:
                command.append(piece)
        return command

    def scale_lr(self, batch_size: int) -> float:
        """Scale the learning rate for the given batch size.

        Assumes that `self.default_bs` and `self.default_lr` were given.
        Then, `self.default_lr` is scaled for the given `batch_size` using
        square root scaling for adaptive optimizers (e.g. Adam, Adadelta,
        AdamW) and linear scaling for others (e.g. SGD).
        """
        assert self.default_bs, "You must provide default_bs to scale LR."
        assert self.default_lr, "You must provide default_lr to scale LR."

        optimizer = self.optimizer.lower()
        if optimizer in ["adam", "adadelta", "adamw"]:
            scaler = SquareRootScaler(bs=self.default_bs, lr=self.default_lr)
            return scaler.compute_lr(batch_size)
        if optimizer in ["sgd"]:
            scaler = LinearScaler(bs=self.default_bs, lr=self.default_lr)
            return scaler.compute_lr(batch_size)
        raise NotImplementedError(f"LR scaling for {self.optimizer} is not supported.")



# Copyright (C) 2022 Jae-Won Chung <jwnchung@umich.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the `ZeusMaster` class, which drives batch size optimization and training."""

from __future__ import annotations

import json
import os
import pprint
import subprocess
from copy import deepcopy
from pathlib import Path
from time import localtime, sleep, strftime

import numpy as np
import pynvml
import torch

from zeus.analyze import HistoryEntry
from zeus.job import Job
from zeus.policy import BatchSizeOptimizer
from zeus.util import zeus_cost


class ZeusMaster:
    """Drives Zeus across multiple recurrences of a job.

    The main purpose of `ZeusMaster` is to launch training scripts integrated
    with [`ZeusDataLoader`][zeus.run.ZeusDataLoader], controlling it by setting
    environment variables. For the environment variables, see the
    [`run_job`][zeus.run.ZeusMaster.run_job] method as well as
    [`ZeusDataLoader`][zeus.run.ZeusDataLoader]'s class docstring.

    The optimal batch size is searched for and exploited using the
    [`BatchSizeOptimizer`][zeus.policy.BatchSizeOptimizer] object passed in
    through the constructor.
    """

    def __init__(
        self,
        batch_size_optimizer: BatchSizeOptimizer,
        log_base: str,
        monitor_path: str,
        seed: int = 123456,
        observer_mode: bool = False,
        profile_warmup_iters: int = 10,
        profile_measure_iters: int = 40,
    ) -> None:
        """Initialize the master.

        Args:
            batch_size_optimizer: The user is expected to construct the
                [`BatchSizeOptimizer`][zeus.policy.BatchSizeOptimizer] with the desired
                policy and pass it into the master class.
            log_base: Absolute path where logs will be stored. A separate directory
                will be created inside, whose name is determined by the job and current time.
            monitor_path: Absolute path to the power monitor binary.
            seed: The random seed. Every invocation of the [`run`][zeus.run.ZeusMaster.run]
                method in this class is deterministic given the random seed, because the
                internal RNG states are deepcopied before servicing jobs.
            observer_mode: When Observer Mode is on, the maximum power limit is
                always used instead of the optimal power limit. However, internal time and
                energy accounting will be done as if the cost-optimal power limit is used.
            profile_warmup_iters: Number of iterations to warm up on a specific power limit.
                This is passed to the [`ZeusDataLoader`][zeus.run.ZeusDataLoader].
            profile_measure_iters: Number of iterations to measure on a specific power limit.
                This is passed to the [`ZeusDataLoader`][zeus.run.ZeusDataLoader].

        """
        # Knob optimizers.
        self.bso = batch_size_optimizer

        # Check if monitor_path is absolute.
        # This is needed since we may change the cwd based on the job's workdir.
        if not Path(monitor_path).is_absolute():
            raise ValueError("monitor_path must be specified as an absolute path.")

        # Log base directory.
        # Needs to be absolute because the training job script may have a different
        # current working directory (when job.workdir is not None).
        if not Path(log_base).is_absolute():
            raise ValueError("log_base must be specified as an absolute path.")
        os.makedirs(log_base, exist_ok=True)
        self.log_base = log_base

        # Save arguments.
        self.seed = seed
        self.monitor_path = monitor_path
        self.observer_mode = observer_mode
        self.profile_warmup_iters = profile_warmup_iters
        self.profile_measure_iters = profile_measure_iters

        # Query the max power limit of the GPU.
        pynvml.nvmlInit()
        handle = pynvml.nvmlDeviceGetHandleByIndex(0)
        minmax = pynvml.nvmlDeviceGetPowerManagementLimitConstraints(handle)  # unit: mW
        self.max_pl = minmax[1] // 1000  # unit: W
        print(
            f"[Zeus Master] Max power limit of {pynvml.nvmlDeviceGetName(handle)}: {self.max_pl}W"
        )
        pynvml.nvmlShutdown()

    def build_logdir(
        self,
        job: Job,
        num_recurrence: int,
        eta_knob: float,
        beta_knob: float,
        exist_ok: bool = True,
    ) -> str:
        r"""Build the `ZEUS_LOG_DIR` string and create the directory.

        Args:
            job: Job to run.
            num_recurrence: The total number of recurrences.
            eta_knob: $\eta$ used in the cost metric.
                $\textrm{cost} = \eta \cdot \textrm{ETA} + (1 - \eta) \cdot \textrm{MaxPower} \cdot \textrm{TTA}$
            beta_knob: `beta_knob * min_cost` is the early stopping cost threshold.
                Set to `np.inf` to disable early stopping.
            exist_ok: Passed to `os.makedirs`. If `False`, will err if the directory
                already exists.
        """
        now = strftime("%Y%m%d%H%M%s", localtime())
        logdir = (
            job.to_logdir() + f"+x{num_recurrence}+eta{eta_knob}+beta{beta_knob}+{now}"
        )
        logdir = f"{self.log_base}/{logdir}"
        os.makedirs(logdir, exist_ok=exist_ok)
        return logdir

    def run_job(
        self,
        job: Job,
        batch_size: int,
        learning_rate: float,
        seed: int,
        logdir: str,
        rec_i: int,
        tries: int,
        eta_knob: float,
        cost_ub: float,
    ) -> tuple[float, float, bool]:
        r"""Launch the training job.

        Args:
            job: The job to run.
            batch_size: The batch size to use.
            learning_rate: The learning rate to use, scaled based on `batch_size`.
            seed: The random seed to use for training.
            logdir: Directory to store log files in.
            rec_i: Recurrence number of this run of the job.
            tries: Retry number of this recurrence of the job.
            eta_knob: $\eta$ used in the cost metric.
                $\textrm{cost} = \eta \cdot \textrm{ETA} + (1 - \eta) \cdot \textrm{MaxPower} \cdot \textrm{TTA}$
            cost_ub: Cost upper bound. The job is terminated when the next epoch is going
                to exceed the cost upper bound.

        Returns:
            A tuple of energy consumption, time consumption, and whether the job reached the target metric.
        """
        # Generate job command
        command = job.gen_command(batch_size, learning_rate, seed, rec_i)

        # Set environment variables
        job_id = f"rec{rec_i:02d}+try{tries:02d}"
        zeus_env = dict(
            ZEUS_LOG_DIR=logdir,
            ZEUS_JOB_ID=job_id,
            ZEUS_COST_THRESH="inf" if cost_ub == np.inf else str(cost_ub),
            ZEUS_ETA_KNOB=str(eta_knob),
            ZEUS_TARGET_METRIC=str(job.target_metric),
            ZEUS_MONITOR_PATH=self.monitor_path,
            ZEUS_PROFILE_PARAMS=f"{self.profile_warmup_iters},{self.profile_measure_iters}",
            ZEUS_USE_OPTIMAL_PL=str(not self.observer_mode),
        )
        env = deepcopy(os.environ)
        env.update(zeus_env)

        # Training script output captured by the master.
        job_output = f"{logdir}/{job_id}.train.log"

        # Training stats (energy, time, reached, end_epoch) written by ZeusDataLoader.
        # This file being found means that the training job is done.
        train_json = Path(f"{logdir}/{job_id}+bs{batch_size}.train.json")

        # Reporting
        print(f"[run job] Launching job with BS {batch_size}:")
        print(f"[run job] {zeus_env=}")
        if job.workdir is not None:
            print(f"[run job] cwd={job.workdir}")
        print(f"[run job] {command=}")
        print(f"[run job] {cost_ub=}")
        print(f"[run job] Job output logged to '{job_output}'")

        # Run the job.
        with open(job_output, "w") as f:
            # Launch subprocess.
            # stderr is redirected to stdout, and stdout to the job_output file.
            proc = subprocess.Popen(
                command,
                cwd=job.workdir,
                stderr=subprocess.STDOUT,
                stdout=f,
            )

            # Check if training is done.
            with open(job_output, "r") as jobf:
                while proc.poll() is None:
                    print(jobf.read(), end="")
                    sleep(1.0)

                # Print out the rest of the script output.
                f.flush()
                print(jobf.read())

                # Report exitcode.
                exitcode = proc.poll()
                print(f"[run job] Job terminated with exit code {exitcode}.")

            # `train_json` must exist at this point.
            if not train_json.exists():
                raise RuntimeError(f"{train_json} does not exist.")

        # Read `train_json` for the training stats.
        with open(train_json, "r") as f:
            stats = json.load(f)
            print(f"[run job] {stats=}")

        # Casting
        if not isinstance(stats["reached"], bool):
            stats["reached"] = stats["reaached"].lower() == "true"

        return float(stats["energy"]), float(stats["time"]), stats["reached"]

    def run(
        self,
        job: Job,
        num_recurrence: int,
        batch_sizes: list[int],
        beta_knob: float,
        eta_knob: float,
    ) -> list[HistoryEntry]:
        r"""Run a job that sequentially recurs without overlap.

        Args:
            job: The job to run.
            num_recurrence: How many times the job recurs.
            batch_sizes: List of feasible batch sizes.
            beta_knob: `beta_knob * min_eta` is the early stopping cost threshold.
                Set to `np.inf` to disable early stopping.
            eta_knob: $\eta$ used in the cost metric.
                $\textrm{cost} = \eta \cdot \textrm{ETA} + (1 - \eta) \cdot \textrm{MaxPower} \cdot \textrm{TTA}$

        Returns:
            A list of [`HistoryEntry`][zeus.analyze.HistoryEntry] objects for each job run.
        """
        # Sanity checks
        if job.default_bs is None:
            raise ValueError("You must provide a default batch size for the job.")
        if job.command is None:
            raise ValueError("You must provide a command format string for the job.")
        if eta_knob < 0.0 or eta_knob > 1.0:
            raise ValueError("eta_knob must be in [0.0, 1.0].")

        print(f"[Zeus Master] {job} x {num_recurrence}")
        print(f"[Zeus Master] Batch sizes: {batch_sizes}")

        # Copy all internal state so that simulation does not modify any
        # internal state and is deterministic w.r.t. the random seed.
        bso = deepcopy(self.bso)
        seed = self.seed

        # ZEUS_LOG_DIR: Where all the logs and files are stored for this run.
        logdir = self.build_logdir(job, num_recurrence, eta_knob, beta_knob)

        # Job history list to return.
        history: list[HistoryEntry] = []

        # Save job history to this file, continuously.
        history_file = f"{logdir}/history.py"

        # beta_knob * min_cost is the early stopping cost threshold.
        min_cost = np.inf

        # Register the job in the BSO.
        bso.register_job(job, batch_sizes)

        # Job recurs.
        for rec_i in range(1, num_recurrence + 1):
            print(f"\n[Zeus Master] Recurrence: {rec_i}")

            # The retrying loop. Retry until convergence.
            cost_acc = 0.0
            for tries in range(1, 21):
                # Fetch the knobs to use.
                bs = bso.predict(job)

                # Scale the learning rate.
                lr = job.scale_lr(bs)

                # Launch the job.
                # Power profiling and optimization is done entirely by the ZeusDataLoader.
                # Early stops based on cost_ub.
                energy, time, reached = self.run_job(
                    job=job,
                    batch_size=bs,
                    learning_rate=lr,
                    seed=seed,
                    logdir=logdir,
                    rec_i=rec_i,
                    tries=tries,
                    eta_knob=eta_knob,
                    cost_ub=beta_knob * min_cost,
                )

                # The random seed will be unique for each run, but still jobs will be
                # deterministic w.r.t. each call to `run`.
                seed += 1

                # Compute the cost of this try.
                num_gpus = torch.cuda.device_count()
                cost = zeus_cost(energy, time, eta_knob, self.max_pl * num_gpus)
                print(f"[Zeus Master] {cost=}")

                # Accumulate the cost to track the total cost of this recurrence.
                cost_acc += cost

                # Provide feedback to the BSO.
                bso.observe(job, bs, cost, reached)

                # Record history for visualization.
                history.append(HistoryEntry(bs, None, energy, reached, time))
                with open(history_file, "w") as f:
                    # Intended use:
                    #
                    # ```python
                    # from zeus.analyze import HistoryEntry
                    # history = eval(open(history_file).read())
                    # ```
                    f.write(pprint.pformat(history) + "\n")

                # Reached the target metric. Go to next recurrence.
                if reached:
                    print(
                        "\n[Zeus Master] Reached target metric in "
                        f"{tries} {'try' if tries == 1 else 'tries'}."
                    )
                    # Track the minimum cost.
                    if min_cost > cost_acc:
                        print(
                            f"\n[Zeus Master] Minimum cost updated from {min_cost} to {cost_acc}."
                        )
                        min_cost = cost_acc
                    break
                # Didn't reach the target metric.
                # We assume that the default BS (set by the user) will converge.
                if rec_i == 1:
                    raise RuntimeError(
                        f"The default batch size {job.default_bs} did not converge."
                    )
            else:
                print(
                    "\n[Zeus Master] Job did not reach the target metric in 20 trials!"
                )
                raise RuntimeError("Unreachable target metric.")

        print(f"[Zeus Master]\n{history}")

        return history



# Copyright (C) 2022 Jae-Won Chung <jwnchung@umich.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for logging."""

import sys
from pathlib import Path


class FileAndConsole:
    """Like tee, but for Python prints."""

    def __init__(self, filepath: Path) -> None:
        """Initialize the object."""
        self.file = open(filepath, "w")
        self.stdout = sys.stdout

    def write(self, message):
        """Write message."""
        self.file.write(message)
        self.stdout.write(message)
        self.file.flush()
        self.stdout.flush()

    def flush(self):
        """Flush both log file and stdout."""
        self.file.flush()
        self.stdout.flush()



# Copyright (C) 2022 Jae-Won Chung <jwnchung@umich.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tools for checking stuff."""

from __future__ import annotations

import os
import unittest
from contextlib import contextmanager
from typing import Any, Type, TypeVar, cast

T = TypeVar("T")


def get_env(name: str, valtype: Type[T], default: T | None = None) -> T:
    """Fetch an environment variable and cast it to the given type."""
    try:
        if valtype == bool:
            val = os.environ[name].lower()
            if val not in ["true", "false"]:
                raise RuntimeError(
                    f"Strange boolean environment variable value '{val}'"
                )
            return cast(T, val == "true")
        return valtype(os.environ[name])
    except KeyError:
        if default is not None:
            return default
        raise RuntimeError(f"Zeus requires the environment variable '{name}'") from None


@contextmanager
def _temp_env(key: str, value: Any):
    if key in os.environ:
        old = os.environ[key]
        os.environ[key] = str(value)
        yield key, value
        os.environ[key] = old
    else:
        os.environ[key] = str(value)
        yield key, value
        os.environ.pop(key)


class TestGetEnv(unittest.TestCase):
    """Test get_env."""

    def test_int(self):
        """Test the case when the value is type int."""
        with _temp_env("TEST_INT", 123) as (key, value):
            self.assertEqual(get_env(key, int), value)

    def test_str(self):
        """Test the case when the value is type str."""
        with _temp_env("TEST_STR", "hello world") as (key, value):
            self.assertEqual(get_env(key, str), value)

    def test_bool(self):
        """Test the case when the value is type bool."""
        with _temp_env("TEST_BOOL", "true") as (key, _):
            self.assertEqual(get_env(key, bool), True)
        with _temp_env("TEST_BOOL", "True") as (key, _):
            self.assertEqual(get_env(key, bool), True)
        with _temp_env("TEST_BOOL", "false") as (key, _):
            self.assertEqual(get_env(key, bool), False)
        with _temp_env("TEST_BOOL", "False") as (key, _):
            self.assertEqual(get_env(key, bool), False)
        with _temp_env("TEST_BOOL", "fault") as (key, _):
            with self.assertRaises(RuntimeError):
                get_env(key, bool)

    def test_no_env(self):
        """Test the case when there is no such env var."""
        with self.assertRaises(RuntimeError):
            get_env("TEST_NONE", str)

    def test_default(self):
        """Test the case when there is no such env var but there's a default."""
        self.assertEqual(get_env("TEST_NONE", int, default=123), 123)

    def test_default_ignore(self):
        """Test the case when the default should be ignored when the env var exists."""
        with _temp_env("TEST_DEFAULT", 123) as (key, value):
            self.assertEqual(get_env(key, int, default=321), value)


if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2022 Jae-Won Chung <jwnchung@umich.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Implementations for various optimization policies.

[`JITPowerLimitOptimizer`][zeus.policy.optimizer.JITPowerLimitOptimizer] and
[`PruningGTSBatchSizeOptimizer`][zeus.policy.optimizer.PruningGTSBatchSizeOptimizer]
are the implementations used in Zeus's publication.
"""

from __future__ import annotations

import unittest
from collections import defaultdict
from typing import Generator

import numpy as np

from zeus.job import Job
from zeus.policy.interface import BatchSizeOptimizer, PowerLimitOptimizer
from zeus.policy.mab import GaussianTS


class GTSBatchSizeOptimizer(BatchSizeOptimizer):
    """One Gaussian Thompson Sampling MAB for each job."""

    def __init__(
        self,
        learn_reward_precision: bool,
        reward_precision: float = 0.0,
        prior_mean: float = 0.0,
        prior_precision: float = 0.0,
        num_exploration: int = 1,
        seed: int = 123456,
        verbose: bool = True,
    ) -> None:
        """Initialze the optimizer.

        Refer to the constructor of [`GaussianTS`][zeus.policy.mab.GaussianTS]
        for descriptions of other arguments.

        Args:
            learn_reward_precision: Whether to learn the reward precision of
                each arm as we accumulate observations.
        """
        self.learn_reward_precision = learn_reward_precision
        self.reward_precision = reward_precision
        self.prior_mean = prior_mean
        self.prior_precision = prior_precision
        self.num_exploration = num_exploration
        self.seed = seed
        self.verbose = verbose

        # One MAB for each job.
        self.mabs: dict[Job, GaussianTS] = {}

        # Track the batch size range for each job.
        self.batch_sizes: dict[Job, list[int]] = {}

        # Observation history (batch size, reward) for each job.
        self.history: dict[Job, defaultdict[int, list[float]]] = {}

    @property
    def name(self) -> str:
        """Name of the batch size optimizer."""
        return "GaussianTS BSO"

    def register_job(self, job: Job, batch_sizes: list[int]) -> None:
        """Instantiate a new GaussianTS MAB for the new job."""
        # We do not want to reset the state related to this job if
        # anything already exists.
        if job in self.mabs:
            return
        self.mabs[job] = GaussianTS(
            arms=batch_sizes,
            reward_precision=self.reward_precision,
            prior_mean=self.prior_mean,
            prior_precision=self.prior_precision,
            num_exploration=self.num_exploration,
            seed=self.seed,
            verbose=self.verbose,
        )
        self.batch_sizes[job] = batch_sizes
        self.history[job] = defaultdict(list)
        if self.verbose:
            self._log(f"Registered {job}")

    def predict(self, job: Job) -> int:
        """Return the batch size to use for the job."""
        if self.verbose:
            self._log(f"Prediction for {job}")
        pred = self.mabs[job].predict()
        if self.verbose:
            self._log(f"{job} -> \033[31mBS = {pred}\033[0m")
        return pred

    def observe(
        self, job: Job, batch_size: int, cost: float, converged: bool | None = None
    ) -> None:
        """Learn from the cost of using the given batch size for the job."""
        if batch_size not in self.batch_sizes[job]:
            raise ValueError(f"Unknown batch size '{batch_size}' for {job}.")

        # No normalization needed since we learn a separate bandit for each job.
        reward = -cost

        # Add observation to history.
        self.history[job][batch_size].append(reward)

        # When we're not learning the reward precision, everyting is
        # simple. We can just call `partial_fit` on the job's MAB instance.
        if not self.learn_reward_precision:
            self.mabs[job].fit([batch_size], [reward], reset=False)
            if self.verbose:
                self._log(f"{job} @ {batch_size}: reward = {reward:.2f}")

        # When we're learning the reward precision, we need to
        # 1. re-compute the precision this arm based on the history,
        # 2. update the arm's reward precision
        # 3. and `fit` the new MAB instance on all past data.
        else:
            arm_rewards = np.array(self.history[job][batch_size])
            variance = np.var(arm_rewards)
            # When there is only one observation for the arm, the variance is zero.
            # NOTE: We might still want to have a pre-determined reward precision here
            #       because sampling from an infinite precision Gaussian distribution
            #       always returns the mean (the observation), and it will hamper
            #       exploration in the early stage.
            if variance == 0.0:
                precision = np.inf
            else:
                precision = np.reciprocal(variance)
            mab = self.mabs[job]
            mab.arm_reward_prec[batch_size] = precision
            mab.fit_arm(batch_size, arm_rewards, reset=True)
            self.mabs[job] = mab
            if self.verbose:
                arm_rewards_repr = ", ".join([f"{r:.2f}" for r in arm_rewards])
                self._log(
                    f"{job} @ {batch_size}: "
                    f"arm_rewards = [{arm_rewards_repr}], reward_prec = {precision}"
                )


class PruningExploreManager:
    """Helper class that generates batch sizes to explore and prune."""

    def __init__(
        self,
        batch_sizes: list[int],
        default: int,
        num_pruning_rounds: int = 2,
    ) -> None:
        """Initialze the object.

        Args:
            batch_sizes: The initial set of batch sizes to prune from.
            default: The default batch size (b0) to begin exploration from.
            num_pruning_rounds: How many rounds to run pruning.
        """
        # Sanity checks.
        if default not in batch_sizes:
            raise ValueError(f"Default batch size {default} not in {batch_sizes}.")

        # Save arguments.
        self.batch_sizes = batch_sizes
        self.default = default
        self.num_pruning_rounds = num_pruning_rounds

        # State
        self.expecting = default

        # Generator that returns batch sizes.
        self.gen = self._exploration_engine()

    def _exploration_engine(
        self,
    ) -> Generator[int | None, tuple[int, float, bool], list[int]]:
        """Drive pruning exploration.

        Yields the batch size to be explored.
        The caller should `send` a tuple of (explored batch size, cost, whether reached).
        As a safety measure, the explored batch size must match the most recently yielded
        batch size, and otherwise a `RuntimeError` is raised.
        Finally, when exploration is over, returns a sorted list of batch sizes that
        survived pruning.
        """
        for _ in range(self.num_pruning_rounds):
            # A list of batch sizes that reached the target metric.
            good: list[int] = []

            # We first explore downwards form the default batch size, and then go upwards.
            idx = self.batch_sizes.index(self.default)
            down = sorted(self.batch_sizes[: idx + 1], reverse=True)
            up = sorted(self.batch_sizes[idx + 1 :])

            # We track the best cost because the default batch size is updated to the batch
            # size that performed the best.
            best_cost = np.inf

            for bs_list in [down, up]:
                for bs in bs_list:
                    # We tell the outside world to explore `bs`, and we expect the outside
                    # world to give us back the cost of that `bs`.
                    self.expecting = bs
                    batch_size, cost, reached = yield bs
                    if self.expecting != batch_size:
                        raise RuntimeError(
                            f"PruningExplorationManager: {self.expecting=}, {batch_size=}"
                        )
                    self.expecting = 0

                    # An empty `yield` to not proceed to the next batch size when the caller
                    # `send`s in the results.
                    yield

                    # Only batch sizes that reached the target mteric are good.
                    if reached:
                        if best_cost > cost:
                            best_cost = cost
                            self.default = bs
                        good.append(bs)
                    # If the batch size did not reach the target metric, `break`ing here will
                    # allow us to move on to either the next direction of exploration (upwards)
                    # or end this round of pruning exploration.
                    else:
                        break

            self.expecting = 0
            self.batch_sizes = sorted(good)

        return sorted(self.batch_sizes)

    def next_batch_size(self) -> int:
        """Return the next batch size to explore.

        Raises `StopIteration` when pruning exploration phase is over.
        The exception instance contains the final set of batch sizes to consider.
        Access it through `exception.value`.
        """
        batch_size = next(self.gen)
        assert batch_size is not None, "Call order may have been wrong."
        return batch_size

    def report_batch_size_result(
        self, batch_size: int, cost: float, reached: bool
    ) -> None:
        """Report whether the previous batch size reached the target metric.

        Args:
            batch_size: The batch size which this cost observation is from.
            cost: The energy-time cost of running the job with this batch size.
            reached: Whether the job reached the target metric.
        """
        none = self.gen.send((batch_size, cost, reached))
        assert none is None, "Call order may have been wrong."


class TestPruningExploreManager(unittest.TestCase):
    """Unit test class for pruning exploration."""

    batch_sizes: list[int] = [8, 16, 32, 64, 128, 256]

    def run_exploration(
        self,
        manager: PruningExploreManager,
        exploration: list[tuple[int, float, bool]],
        result: list[int],
    ) -> None:
        """Drive the pruning explore manager and check results."""
        for bs, cost, reached in exploration:
            self.assertEqual(manager.next_batch_size(), bs)
            manager.report_batch_size_result(bs, cost, reached)
        with self.assertRaises(StopIteration) as raised:
            manager.next_batch_size()
        self.assertEqual(raised.exception.value, result)

    def test_normal(self):
        """Test a typical case."""
        manager = PruningExploreManager(self.batch_sizes, 128)
        exploration = [
            (128, 10.0, True),
            (64, 9.0, True),
            (32, 8.0, True),
            (16, 12.0, True),
            (8, 21.0, False),
            (256, 15.0, True),
            (32, 8.0, True),
            (16, 12.0, False),
            (64, 9.0, True),
            (128, 10.0, True),
            (256, 17.0, False),
        ]
        result = [32, 64, 128]
        self.run_exploration(manager, exploration, result)

    def test_default_is_largest(self):
        """Test the case when the default batch size is the largest one."""
        manager = PruningExploreManager(self.batch_sizes, 256)
        exploration = [
            (256, 7.0, True),
            (128, 8.0, True),
            (64, 9.0, True),
            (32, 13.0, True),
            (16, 22.0, False),
            (256, 8.0, True),
            (128, 8.5, True),
            (64, 9.0, True),
            (32, 12.0, True),
        ]
        result = [32, 64, 128, 256]
        self.run_exploration(manager, exploration, result)

    def test_default_is_smallest(self):
        """Test the case when the default batch size is the smallest one."""
        manager = PruningExploreManager(self.batch_sizes, 8)
        exploration = [
            (8, 10.0, True),
            (16, 17.0, True),
            (32, 20.0, True),
            (64, 25.0, False),
            (8, 10.0, True),
            (16, 21.0, False),
        ]
        result = [8]
        self.run_exploration(manager, exploration, result)

    def test_all_converge(self):
        """Test the case when every batch size converges."""
        manager = PruningExploreManager(self.batch_sizes, 64)
        exploration = [
            (64, 10.0, True),
            (32, 8.0, True),
            (16, 12.0, True),
            (8, 15.0, True),
            (128, 12.0, True),
            (256, 13.0, True),
            (32, 7.0, True),
            (16, 10.0, True),
            (8, 15.0, True),
            (64, 10.0, True),
            (128, 12.0, True),
            (256, 13.0, True),
        ]
        result = self.batch_sizes
        self.run_exploration(manager, exploration, result)

    def test_every_bs_is_bs(self):
        """Test the case when every batch size other than the default fail to converge."""
        manager = PruningExploreManager(self.batch_sizes, 64)
        exploration = [
            (64, 10.0, True),
            (32, 22.0, False),
            (128, 25.0, False),
            (64, 9.0, True),
        ]
        result = [64]
        self.run_exploration(manager, exploration, result)


class PruningGTSBatchSizeOptimizer(BatchSizeOptimizer):
    """One Gaussian Thompson Sampling MAB for each job with double pruning exploration."""

    def __init__(
        self,
        prior_mean: float = 0.0,
        prior_precision: float = 0.0,
        window_size: int = 0,
        concurrency: bool = False,
        seed: int = 123456,
        verbose: bool = True,
    ) -> None:
        """Initialze the optimizer.

        Refer to the constructor of [`GaussianTS`][zeus.policy.mab.GaussianTS]
        for descriptions of other arguments.

        Args:
            window_size: Size of the window for the MAB (for drift handling).
            concurrency: Whether to support concurrent job submissions.
        """
        self.prior_mean = prior_mean
        self.prior_precision = prior_precision
        self.window_size = window_size
        self.concurrency = concurrency
        self.seed = seed
        self.verbose = verbose

        # One MAB for each job.
        self.mabs: dict[Job, GaussianTS] = {}

        # One PruningExplorationManager for each job.
        self.exp_manager: dict[Job, PruningExploreManager] = {}

        # Observation history (batch size, reward) for each job.
        self.history: dict[Job, list[tuple[int, float]]] = {}

    @property
    def name(self) -> str:
        """Name of the batch size optimizer."""
        return "Pruning GaussianTS BSO"

    def register_job(self, job: Job, batch_sizes: list[int]) -> None:
        """Register the job."""
        # Sanity checks.
        if job.default_bs is None:
            raise ValueError(f"Default BS not specified for {job}.")
        if not batch_sizes:
            raise ValueError(f"Batch size list for {job} is empty.")

        # Set internal states.
        self.exp_manager[job] = PruningExploreManager(
            sorted(batch_sizes), job.default_bs
        )
        self.history[job] = []
        if self.verbose:
            self._log(f"Registered {job}")

    def predict(self, job: Job) -> int:
        """Return the batch size to use for the job."""
        # Try to see if the exploration manager has something.
        try:
            batch_size = self.exp_manager[job].next_batch_size()
            if self.verbose:
                self._log(f"{job} in pruning stage -> \033[31mBS = {batch_size}\033[0m")
        except StopIteration as exp:
            # Pruning stage is over.
            if job not in self.mabs:
                self._construct_mab(job, exp.value)
            batch_size = self.mabs[job].predict()
            if self.verbose:
                self._log(
                    f"{job} in Thompson Sampling stage -> \033[31mBS = {batch_size}\033[0m"
                )

        return batch_size

    def observe(
        self, job: Job, batch_size: int, cost: float, converged: bool | None = None
    ) -> None:
        """Learn from the cost of using the given batch size for the job."""
        # Add observation to history.
        self.history[job].append((batch_size, -cost))

        # We're in Thompson Sampling stage.
        if job in self.mabs:
            # Since we're learning the reward precision, we need to
            # 1. re-compute the precision of this arm based on the reward history,
            # 2. update the arm's reward precision
            # 3. and `fit` the new MAB instance on all the reward history.
            # Note that `arm_rewards` always has more than one entry (and hence a
            # non-zero variance) because we've been through pruning exploration.
            arm_rewards = np.array(self._get_history_for_bs(job, batch_size))
            precision = np.reciprocal(np.var(arm_rewards))
            mab = self.mabs[job]
            mab.arm_reward_prec[batch_size] = precision
            mab.fit_arm(batch_size, arm_rewards, reset=True)
            if self.verbose:
                arm_rewards_repr = ", ".join([f"{r:.2f}" for r in arm_rewards])
                self._log(
                    f"{job} @ {batch_size}: "
                    f"arm_rewards = [{arm_rewards_repr}], reward_prec = {precision}"
                )

        # We're in pruning stage.
        else:
            assert converged is not None
            # Log before we potentially error out.
            if self.verbose:
                self._log(
                    f"{job} in pruning stage, expecting BS {self.exp_manager[job].expecting}."
                    f" Current BS {batch_size} that did {'not ' * converged}converge."
                )

            # If we don't support concurrency, we can just pass the results to the
            # exploration manager, and the manager will err if the order of batch sizes
            # is screwed up.
            if not self.concurrency:
                self.exp_manager[job].report_batch_size_result(
                    batch_size, cost, converged
                )
                return

            # If we are supporting concurrency, there's a subtle issue.
            # Pruning exploration demands a specific order of trying out a batch size
            # and receiving the results (cost and whether reached). This breaks in the
            # following situation, for example:
            # 1. Job with BS 32 that is part of pruning exploration starts.
            # 2. Concurrent job comes in, and we launch it with the best known BS 64.
            # 3. Job with BS 64 finishes first, and calls bso.observe with BS 64.
            # This breaks the observation order assumption of PruningExplorationManager.
            # Thus we check whether the current batch size is the one expected by
            # PruningExplorationManager, and then only if so, call bso.observe.
            # Otherwise, we silently insert the cost observation into the bso's history
            # (first line of this method) and don't touch the PruningExplorationManager.
            if self.exp_manager[job].expecting == batch_size:
                self.exp_manager[job].report_batch_size_result(
                    batch_size, cost, converged
                )

    def _get_history_for_bs(self, job: Job, batch_size: int) -> list[float]:
        """Return the windowed history for the given job's batch size."""
        history = self.history[job]
        rewards = []
        # Collect rewards starting from the most recent ones and backwards.
        for bs, reward in reversed(history):
            if bs == batch_size:
                rewards.append(reward)
                if len(rewards) == self.window_size:
                    break
        # There's no need to return this in time order, but just in case.
        return list(reversed(rewards))

    def _construct_mab(self, job: Job, batch_sizes: list[int]) -> None:
        """When exploration is over, this method is called to construct and learn GTS."""
        # Sanity check.
        if not batch_sizes:
            raise ValueError(
                "Empty batch size set when constructing MAB. "
                "Probably all batch sizes have been pruned."
            )

        if self.verbose:
            self._log(f"Construct MAB for {job} with arms {batch_sizes}")

        mab = GaussianTS(
            arms=batch_sizes,  # The MAB only has "good" arms.
            reward_precision=0.0,
            prior_mean=self.prior_mean,
            prior_precision=self.prior_precision,
            num_exploration=2,
            seed=self.seed,
            verbose=self.verbose,
        )
        # Fit the arm for each good batch size.
        for batch_size in self.exp_manager[job].batch_sizes:
            arm_rewards = np.array(self._get_history_for_bs(job, batch_size))
            assert (
                len(arm_rewards) >= 2
            ), f"Number of observations for {batch_size} is {len(arm_rewards)}."
            mab.arm_reward_prec[batch_size] = np.reciprocal(np.var(arm_rewards))
            mab.fit_arm(batch_size, arm_rewards, reset=True)
        # Save the MAB.
        self.mabs[job] = mab


class JITPowerLimitOptimizer(PowerLimitOptimizer):
    """Returns the best power limit to use for the job & batch size."""

    def __init__(self, verbose: bool = True) -> None:
        """Initialize the object."""
        self.verbose = verbose

        self.best_pl: defaultdict[Job, dict[int, int]] = defaultdict(dict)
        self.best_cost: defaultdict[Job, dict[int, float]] = defaultdict(dict)
        self.observe_count: defaultdict[Job, defaultdict[int, int]] = defaultdict(
            lambda: defaultdict(int)
        )

    @property
    def name(self) -> str:
        """Name of the power limit optimizer."""
        return "JITPSO"

    def predict(self, job: Job, batch_size: int) -> int | None:
        """Return the best power limit for the job, or None if unknown."""
        pred = self.best_pl[job].get(batch_size)
        if self.verbose:
            self._log(
                f"{job} @ {batch_size} -> \033[31mPL = "
                f"{'needs profiling' if pred is None else str(pred) + 'W'}\033[0m"
            )
        return pred

    def observe(self, job: Job, batch_size: int, power_limit: int, cost: float) -> None:
        """Learn from the cost of using the given knobs for the job."""
        self.observe_count[job][batch_size] += 1
        prev_best_cost = self.best_cost[job].get(batch_size)
        if prev_best_cost is None or prev_best_cost > cost:
            self.best_pl[job][batch_size] = power_limit
            self.best_cost[job][batch_size] = cost


if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2022 Jae-Won Chung <jwnchung@umich.edu>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

setup(
    name="zeus-ml",
    version="0.2.1",
    description="An Energy Optimization Framework for DNN Training",
    long_description="# Zeus: An Energy Optimization Framework for DNN Training\n",
    long_description_content_type="text/markdown",
    url="https://github.com/SymbioticLab/Zeus",
    author="Jae-Won Chung",
    author_email="jwnchung@umich.edu",
    license="Apache-2.0",
    classifiers=[
        "Environment :: GPU :: NVIDIA CUDA",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords=["deep-learning", "power", "energy", "mlsys"],
    project_urls={
        "Documentation": "https://ml.energy/zeus",
    },
    packages=find_packages("."),
    install_requires=[
        "torch",
        "numpy",
        "pandas==1.4.2",
        "scikit-learn",
        "pynvml",
    ],
    python_requires=">=3.8",
    extras_require={
        "dev": ["pylint==2.14.5", "black==22.6.0", "pydocstyle==6.1.1"],
    }
)
